import torch
from typing import Union, Optional, Callable, cast
import pandas as pd
import numpy as np
from .structural_model import StructuralModel
from .utils import device, smoke_test
from tqdm.notebook import tqdm
from scipy.optimize import minimize


class NlmeModel:
    def __init__(
        self,
        structural_model: StructuralModel,
        patients_df: pd.DataFrame,
        init_log_MI: dict[str, float],
        init_PDU: dict[str, dict[str, float]],
        init_res_var: list[float],
        covariate_map: Optional[dict[str, dict[str, dict[str, str | float]]]] = None,
        error_model_type: str = "additive",
        pred_var_threshold: float = 1e-2,
        num_chains: int = 3,
        constraints: Optional[dict[str, dict]] = None,
    ):
        """Create a non-linear mixed effects model

        Using a structural model (simulation model) and a covariate structure, create a non-linear mixed effects model, to be used in PySAEM or another optimizer, or to predict data using a covariance structure.

        Args:
            structural_model (StructuralModel): A simulation model defined via the convenience class StructuralModel
            patients_df (DataFrame): the list of patients to be considered, with potential covariate values listed, and the name of the protocol arm on which the patient was evaluated (optional - if not supplied, `identity` will be used). The `id` column is expected, any additional column will be handled as a covariate
            init_log_MI: for each model intrinsic parameter, provide an initial value (log)
            init_PDU: for each patient descriptor unknown parameter, provide an initial mean and sd of the log
            init_res_var: for each model output, provide an initial residual variance
            covariate_map (optional[dict]): for each PDU, the list of covariates that affect it - each associated with a covariation coefficient (to be calibrated)
            Example
                {"pdu_name":
                    {"covariate_name":
                        {"coef": "coef_name", "value": initial_value}
                    }
                }
            error_model_type (str): either `additive` or `proportional` error model
            pred_var_threshold (float): Threshold of predictive variance that will issue a warning. Default 1e-2.
            num_chains (int): Number of parallel chains to track for each patient
            constraints (dict): Box constraints for each random parameter. Should be a map that associates `pdu_name: {low: value|None, high: value|None}`. Any unspecified constraint will default to `{low: 0, high: Inf}` and result in exponential transformation of the associated gaussian parameter.
        """
        self.structural_model: StructuralModel = structural_model
        self.pred_var_threshold = pred_var_threshold

        self.MI_names: list[str] = list(init_log_MI.keys())
        self.nb_MI: int = len(self.MI_names)
        self.init_log_MI = torch.tensor([val for _, val in init_log_MI.items()]).to(
            device
        )
        self.PDU_names: list[str] = list(init_PDU.keys())
        self.nb_PDU: int = len(self.PDU_names)

        if set(self.MI_names) & set(self.PDU_names):
            raise ValueError(
                f"Overlapping model intrinsic and PDU descriptors:{(set(self.MI_names) & set(self.PDU_names))}"
            )

        self.patients_df: pd.DataFrame = patients_df.drop_duplicates()
        self.patients: list[str | int] = self.patients_df["id"].unique().tolist()
        self.nb_patients: int = len(self.patients)
        covariate_columns = self.patients_df.columns.to_list()
        if "protocol_arm" not in covariate_columns:
            self.patients_df["protocol_arm"] = "identity"

        additional_columns: list[str] = self.patients_df.drop(
            ["id", "protocol_arm"], axis=1
        ).columns.tolist()

        init_betas_list: list = []
        if covariate_map is None:
            print(
                f"No covariate map provided. All additional columns in `patients_df` will be handled as known descriptors: {additional_columns}"
            )
            self.covariate_map = None
            self.covariate_names = []
            self.covariate_coeffs_names = []
            self.nb_covariates = 0
            self.population_betas_names = self.PDU_names
            init_betas_list = [val["mean"] for _, val in init_PDU.items()]
            self.PDK_names = additional_columns
            self.nb_PDK = len(self.PDK_names)
        else:
            self.covariate_map = covariate_map
            self.population_betas_names: list = []
            covariate_set = set()
            covariate_coeffs_set = set()
            pdk_names = set(additional_columns)
            for PDU_name in self.PDU_names:
                self.population_betas_names.append(PDU_name)
                init_betas_list.append(init_PDU[PDU_name]["mean"])
                if PDU_name not in covariate_map:
                    raise ValueError(
                        f"No covariate map listed for {PDU_name}. Add an empty set if it has no covariate."
                    )
                for covariate, coef in self.covariate_map[PDU_name].items():
                    if covariate not in additional_columns:
                        raise ValueError(
                            f"Covariate appears in the map but not in the patient set: {covariate}"
                        )
                    if covariate is not None:
                        covariate_set.add(covariate)
                        if covariate in pdk_names:
                            pdk_names.remove(covariate)
                        coef_name = coef["coef"]
                        covariate_coeffs_set.add(coef_name)
                        coef_val = coef["value"]
                        self.population_betas_names.append(coef_name)
                        init_betas_list.append(coef_val)
            self.covariate_names = list(covariate_set)
            self.covariate_coeffs_names = list(covariate_coeffs_set)
            self.nb_covariates = len(self.covariate_names)
            self.PDK_names = list(pdk_names)
            self.nb_PDK = len(self.PDK_names)

        print(f"Successfully loaded {self.nb_covariates} covariates:")
        print(self.covariate_names)
        if self.nb_PDK > 0:
            self.patients_pdk = {}
            for patient in self.patients:
                row = self.patients_df.loc[
                    self.patients_df["id"] == patient
                ].drop_duplicates()
                self.patients_pdk.update(
                    {
                        patient: torch.as_tensor(
                            row[self.PDK_names].values, device=device
                        )
                    }
                )
            # Store the full pdk tensor on the device
            self.patients_pdk_full = torch.cat(
                [self.patients_pdk[ind] for ind in self.patients]
            ).to(device)
            print(f"Successfully loaded {self.nb_PDK} known descriptors:")
            print(self.PDK_names)
        else:
            # Create an empty pdk tensor
            self.patients_pdk_full = torch.empty((self.nb_patients, 0), device=device)

        if set(self.PDK_names + self.PDU_names + self.MI_names) != set(
            self.structural_model.parameter_names
        ):
            raise ValueError(
                f"Non-matching descriptor set and structural model parameter set:\nPDK, PDU, MI: {set(self.PDK_names + self.PDU_names + self.MI_names)}\nStructural model descriptors: {set(self.structural_model.parameter_names)}"
            )

        self.descriptors: list[str] = self.PDK_names + self.PDU_names + self.MI_names
        self.nb_descriptors: int = len(self.descriptors)
        # Assume that the descriptors will always be provided to the model in the following order:
        #   PDK, PDU, MI
        self.model_input_to_descriptor = torch.as_tensor(
            [
                self.descriptors.index(param)
                for param in self.structural_model.parameter_names
            ],
            device=device,
        ).long()
        self.initial_betas = torch.as_tensor(init_betas_list, device=device)
        self.nb_betas: int = len(self.population_betas_names)
        self.outputs_names: list[str] = self.structural_model.output_names
        self.nb_outputs: int = self.structural_model.nb_outputs
        self.error_model_type: str = error_model_type
        self.init_res_var = torch.as_tensor(init_res_var, device=device)
        # Initialize the covariance matrix with diagonal entries
        self.init_omega = torch.diag(
            torch.as_tensor(
                [float(init_PDU[pdu]["sd"]) for pdu in self.PDU_names], device=device
            )
        )

        # Assemble the list of design matrices from the covariance structure
        self.design_matrices = self._create_all_design_matrices()
        # Store the full design matrix on the device
        # Size (nb_patients, nb_PDU, nb_betas)
        self.full_design_matrix = torch.stack(
            [self.design_matrices[p] for p in self.patients]
        ).to(device)

        # Initiate the NLME parameters
        self.num_chains = num_chains
        self.update_current_parameters(
            omega=self.init_omega,
            beta=self.initial_betas,
            log_mi=self.init_log_MI,
            res_var=self.init_res_var,
        )

        self.pdu_transforms = {"exp": [], "sigmoid": []}
        self.pdu_shift = torch.zeros(1, 1, self.nb_PDU)
        self.pdu_scale = torch.ones(1, 1, self.nb_PDU)

        self.mi_transforms = {"exp": [], "sigmoid": []}
        self.mi_shift = torch.zeros(self.nb_MI)
        self.mi_scale = torch.ones(self.nb_MI)

        if constraints is not None:
            for pdu_index, pdu in enumerate(self.PDU_names):
                if pdu in constraints:
                    # Constraints are specified for this PDU
                    bounds = constraints[pdu]
                    if bounds["low"] is not None:
                        # A lower bound is specified. Include it in the shift
                        self.pdu_shift[0, 0, pdu_index] = bounds["low"]
                        # If no shift is specified, a 0 shift is kept
                    if bounds["high"] is not None:
                        # A high bound is specified: need to include the PDU in the sigmoid transformation group
                        self.pdu_transforms["sigmoid"].append(pdu_index)
                        # Compute the scaling factor
                        self.pdu_scale[0, 0, pdu_index] = (
                            bounds["high"] - self.pdu_shift[0, 0, pdu_index]
                        )
                    else:
                        # No high bound specified -> the exp transformation will be used
                        self.pdu_transforms["exp"].append(pdu_index)
                else:
                    # No constraints specified for this pdu -> the 0 shift and exp transformation will be used
                    self.pdu_transforms["exp"].append(pdu_index)
            print(
                "Successfully loaded the following PDU constraints and transformations:"
            )
            print(self.pdu_transforms)
            print(f"Shift: {self.pdu_shift[0, 0, :]}, scale: {self.pdu_scale[0, 0, :]}")
            for mi_index, mi in enumerate(self.MI_names):
                if mi in constraints:
                    # Constraints are specified for this MI
                    bounds = constraints[mi]
                    if bounds["low"] is not None:
                        # A lower bound is specified. Include it in the shift
                        self.mi_shift[mi_index] = bounds["low"]
                        # If no shift is specified, a 0 shift is kept
                    if bounds["high"] is not None:
                        # A high bound is specified: need to include the MI in the sigmoid transformation group
                        self.mi_transforms["sigmoid"].append(mi_index)
                        # Compute the scaling factor
                        self.mi_scale[mi_index] = (
                            bounds["high"] - self.mi_shift[mi_index]
                        )
                    else:
                        # No high bound specified -> the exp transformation will be used
                        self.mi_transforms["exp"].append(mi_index)

                else:
                    # No constraints specified for this MI -> the 0 shift and exp transformation will be used
                    self.mi_transforms["exp"].append(mi_index)

            print(
                "Successfully loaded the following MI constraints and transformations:"
            )
            print(self.mi_transforms)
            print(f"Shift: {self.mi_shift}, scale: {self.mi_scale}")
        else:
            # No constraints specified at all: 0 shift and exp transformation for all params
            self.pdu_transforms["exp"] = list(np.arange(self.nb_PDU))
            self.mi_transforms["exp"] = list(np.arange(self.nb_MI))

        self.pdu_transforms = {
            "exp": torch.tensor(
                self.pdu_transforms["exp"], dtype=torch.long, device=device
            ),
            "sigmoid": torch.tensor(
                self.pdu_transforms["sigmoid"], dtype=torch.long, device=device
            ),
        }

        self.mi_transforms = {
            "exp": torch.tensor(
                self.mi_transforms["exp"], dtype=torch.long, device=device
            ),
            "sigmoid": torch.tensor(
                self.mi_transforms["sigmoid"], dtype=torch.long, device=device
            ),
        }
        # Initiate the patient parameters
        self.init_etas = self.sample_etas(self.num_chains, self.nb_patients)
        self.update_eta_samples(self.init_etas)
        # Compute a first naive guess for the patient parameters on all chains
        gaussian_params = self.etas_to_gaussian_params(self.eta_samples_chains)
        physical_params = self.gaussian_to_physical_params(gaussian_params, self.log_MI)
        # Average over the chains to estimate the current patient descriptors
        theta = self.assemble_individual_parameters(physical_params).mean(dim=0)

    def _create_design_matrix(self, covariates: dict[str, float]) -> torch.Tensor:
        """
        Creates the design matrix X_i for a single individual based on the model's covariate map.
        This matrix will be multiplied with population betas so that log(theta_i[PDU]) = X_i @ betas + eta_i.
        """
        design_matrix_X_i = torch.zeros((self.nb_PDU, self.nb_betas), device=device)
        col_idx = 0
        for i, PDU_name in enumerate(self.PDU_names):
            design_matrix_X_i[i, col_idx] = 1.0
            col_idx += 1
            if self.covariate_map is not None:
                for covariate in self.covariate_map[PDU_name].keys():
                    design_matrix_X_i[i, col_idx] = float(covariates[covariate])
                    col_idx += 1
        return design_matrix_X_i

    def _create_all_design_matrices(self) -> dict[Union[str, int], torch.Tensor]:
        """Creates a design matrix for each unique individual based on their covariates, given the in the covariates_df."""
        design_matrices = {}
        if self.nb_covariates == 0:
            for ind_id in self.patients:
                design_matrices[ind_id] = self._create_design_matrix({})
        else:
            for ind_id in self.patients:
                individual_covariates = (
                    self.patients_df[self.patients_df["id"] == ind_id]
                    .iloc[0]
                    .drop("id")
                )
                covariates_dict = individual_covariates.to_dict()
                design_matrices[ind_id] = self._create_design_matrix(covariates_dict)
        return design_matrices

    def add_observations(self, observations_df: pd.DataFrame) -> None:
        """Associate the NLME model with a data frame of observations

        Args:
            observations_df (pd.DataFrame): A data frame of observations, with columns
            - `id`: the patient id. Should be consistent with self.patients_df
            - `time`: the observation time
            - `output_name`
            - `value`
        """
        # Store the raw data frame
        self.observations_df = observations_df
        # Data validation
        input_columns = observations_df.columns.tolist()
        unique_outputs = observations_df["output_name"].unique().tolist()
        if "id" not in input_columns:
            raise ValueError(
                "Provided observation data frame should contain `id` column."
            )
        input_patients = observations_df["id"].unique()
        if set(input_patients) != set(self.patients):
            # Note this check might be unnecessary
            raise ValueError(
                f"Missing observations for the following patients: {set(self.patients) - set(input_patients)}"
            )
        if "time" not in input_columns:
            raise ValueError(
                "Provided observation data frame should contain `time` column."
            )
        if not (set(unique_outputs) <= set(self.outputs_names)):
            raise ValueError(
                f"Unknown model output: {set(unique_outputs) - set(self.outputs_names)}"
            )
        if hasattr(self, "observations_tensors"):
            print(
                "Warning: overriding existing observation data frame for the NLME model"
            )
        if "value" not in input_columns:
            raise ValueError(
                "The provided observations data frame does not contain a `value` column."
            )
        processed_df = observations_df[["id", "output_name", "time", "value"]].merge(
            self.patients_df, how="left", on="id"
        )
        processed_df["task"] = processed_df.apply(
            lambda r: r["output_name"] + "_" + r["protocol_arm"], axis=1
        )
        processed_df["task_index"] = processed_df["task"].apply(
            lambda t: self.structural_model.tasks.index(t)
        )
        processed_df["output_index"] = processed_df["output_name"].apply(
            lambda o: self.structural_model.output_names.index(o)
        )
        global_time_steps = (
            processed_df["time"].drop_duplicates().sort_values().to_list()
        )
        processed_df["time_step_index"] = processed_df["time"].apply(
            lambda t: global_time_steps.index(t)
        )
        self.global_time_steps = torch.as_tensor(global_time_steps, device=device)
        self.nb_global_time_steps = self.global_time_steps.shape[0]
        self.global_time_steps_expanded = (
            self.global_time_steps.unsqueeze(0)
            .unsqueeze(-1)
            .repeat((self.nb_patients, 1, 1))
        )
        # Browse the observed data set and store relevant elements
        self.observations_tensors: dict = {}
        self.n_tot_observations_per_output = torch.zeros(self.nb_outputs, device=device)
        for ind, patient in enumerate(self.patients):
            this_patient = processed_df.loc[processed_df["id"] == patient]

            tasks_indices_np = this_patient["task_index"].values
            tasks_indices = torch.as_tensor(tasks_indices_np, device=device).long()

            outputs_indices_np = this_patient["output_index"].values
            outputs_indices = torch.as_tensor(outputs_indices_np, device=device).long()
            # Add counts of observations to the total per output
            self.n_tot_observations_per_output.scatter_add_(
                0,
                outputs_indices,
                torch.ones_like(outputs_indices, device=device, dtype=torch.float64),
            )

            outputs = torch.as_tensor(this_patient["value"].values, device=device)

            time_steps = torch.as_tensor(this_patient["time"].values, device=device)
            time_step_indices = torch.as_tensor(
                this_patient["time_step_index"].values, device=device
            ).long()
            p_index_repeated = torch.full(
                outputs.shape, ind, dtype=torch.int64, device=device
            )

            self.observations_tensors.update(
                {
                    patient: {
                        "observations": outputs,
                        "time_steps": time_steps,
                        "time_step_indices": time_step_indices,
                        "tasks_indices": tasks_indices,
                        "outputs_indices": outputs_indices,
                        "p_index_repeated": p_index_repeated,
                    }
                }
            )
        self.nb_tot_obs = self.n_tot_observations_per_output.sum()

        # Build the full data set tensors
        self.full_obs_data = torch.cat(
            [self.observations_tensors[p]["observations"] for p in self.patients]
        ).to(device)
        assert self.full_obs_data.shape[0] == self.nb_tot_obs

        # Construct the indexing tensors
        self.observation_to_patient_index = (
            torch.cat(
                [
                    self.observations_tensors[p]["p_index_repeated"]
                    for p in self.patients
                ]
            )
            .long()
            .to(device)
        )
        self.observation_to_timestep_index = (
            torch.cat(
                [
                    self.observations_tensors[p]["time_step_indices"]
                    for p in self.patients
                ]
            )
            .long()
            .to(device)
        )
        self.observation_to_task_index = (
            torch.cat(
                [self.observations_tensors[p]["tasks_indices"] for p in self.patients]
            )
            .long()
            .to(device)
        )
        # Construct a tuple allowing to index a 3D tensor of outputs into a 1D tensor of outputs
        self.prediction_index = (
            self.observation_to_patient_index,
            self.observation_to_timestep_index,
            self.observation_to_task_index,
        )

        self.full_output_indices = (
            torch.cat(
                [self.observations_tensors[p]["outputs_indices"] for p in self.patients]
            )
            .long()
            .to(device)
        )
        self.chunk_sizes: list[int] = [
            self.observations_tensors[p]["observations"].cpu().shape[0]
            for p in self.patients
        ]

    def update_omega(self, omega: torch.Tensor) -> None:
        """Update the covariance matrix of the NLME model and the distribution of random effects."""

        if hasattr(self, "omega_pop"):
            expected_shape = self.omega_pop.shape
        else:
            expected_shape = (self.nb_PDU, self.nb_PDU)
        assert (
            omega.shape == expected_shape
        ), f"Wrong shape in omega update: {omega.shape}, expected: {expected_shape}"

        self.omega_pop = omega
        self.omega_pop_lower_chol = torch.linalg.cholesky(self.omega_pop).to(device)
        self.eta_distribution = torch.distributions.MultivariateNormal(
            loc=torch.zeros(self.nb_PDU, device=device),
            covariance_matrix=self.omega_pop,
        )

    def update_res_var(self, residual_var: torch.Tensor) -> None:
        """Update the residual variance of the NLME model, while ensuring it remains positive."""

        if hasattr(self, "residual_var"):
            expected_shape = self.residual_var.shape
        else:
            expected_shape = (self.nb_outputs,)
        assert (
            residual_var.shape == expected_shape
        ), f"Wrong shape in residual variance update: {residual_var.shape}, expected: {expected_shape}"

        self.residual_var = residual_var.clamp(min=1e-6)

    def update_betas(self, betas: torch.Tensor) -> None:
        """Update the betas of the NLME model."""

        if hasattr(self, "population_betas"):
            expected_shape = self.population_betas.shape
        else:
            expected_shape = (self.nb_betas,)
        assert (
            betas.shape == expected_shape
        ), f"Wrong shape in Betas update: {betas.shape}, expected: {expected_shape}"

        self.population_betas = betas

    def update_log_mi(self, log_MI: torch.Tensor) -> None:
        """Update the model intrinsic parameter values of the NLME model."""

        if hasattr(self, "log_MI"):
            expected_shape = self.log_MI.shape
        else:
            expected_shape = (self.nb_MI,)
        assert (
            log_MI.shape == expected_shape
        ), f"Wrong shape in model intrinsic parameters update: {log_MI.shape}, expected: {expected_shape}"

        self.log_MI = log_MI

    def update_eta_samples(self, eta: torch.Tensor) -> None:
        """Update the model current individual random effect samples."""

        if hasattr(self, "eta_samples_chains"):
            expected_shape = self.eta_samples_chains.shape
        else:
            expected_shape = (self.num_chains, self.nb_patients, self.nb_PDU)
        assert (
            eta.shape == expected_shape
        ), f"Wrong shape in eta samples update: {eta.shape}, expected: {expected_shape}"

        self.eta_samples_chains = eta

    def update_current_parameters(
        self,
        omega: torch.Tensor,
        beta: torch.Tensor,
        log_mi: torch.Tensor,
        res_var: torch.Tensor,
    ):
        """Update or initialize the current population parameter values

        Args:
            omega (torch.Tensor): The random effects covariance matrix
            beta (torch.Tensor): The fixed effects vector (means and covariates)
            log_mi (torch.Tensor): The model intrinsic parameters (log-transformed)
        """
        self.update_omega(omega)
        self.update_betas(beta)
        self.update_log_mi(log_mi)
        self.update_res_var(res_var)

    def sample_etas(self, nb_samples, nb_patients) -> torch.Tensor:
        """Sample individual random effects on all chains from the current estimate of Omega

        Returns:
            torch.Tensor : individual random effects for all patients in the population, one per chain. Size: (num_chains, nb_patients, nb_PDUs)
        """
        etas = self.eta_distribution.sample([nb_samples, nb_patients])
        return etas

    def etas_to_gaussian_params(
        self,
        individual_etas: torch.Tensor,
        ind_ids_for_etas: Optional[list[str | int]] = None,
    ) -> torch.Tensor:
        """Compute individual (gaussian) parameters from random effects chains

        Args:
            individual_etas (torch.Tensor): Individual random effects samples. Size: (num_chains, nb_patients, nb_PDU)
            ind_ids_for_etas (List[Union[str, int]]): List of individual ids corresponding to the sampled etas, used to fetch the design matrices

        Returns:
            torch.Tensor: The individual parameters in gaussian (unconstrained) space. Size: (num_chains, nb_patients, nb_PDU)
        """
        if ind_ids_for_etas is None:
            ind_ids_for_etas = self.patients

        (nb_samples, nb_patients_local, nb_PDU) = individual_etas.shape

        assert (nb_patients_local, nb_PDU) == (
            len(ind_ids_for_etas),
            self.nb_PDU,
        ), f"Wrong shape for individual_etas, expected {len(ind_ids_for_etas), self.nb_PDU}, got {nb_patients_local, nb_PDU}"

        if nb_patients_local == self.nb_patients:
            stacked_design_matrices = self.full_design_matrix
        elif nb_patients_local == 1:
            stacked_design_matrices = self.design_matrices[
                ind_ids_for_etas[0]
            ].unsqueeze(0)
        else:
            raise ValueError(
                f"Unsupported amount of patients requested in `etas_to_gaussian` {nb_patients_local}"
            )

        gaussian_params = (
            stacked_design_matrices.expand(nb_samples, -1, -1, -1)
            @ self.population_betas
            + individual_etas
        )

        return gaussian_params

    @torch.compile
    def gaussian_to_physical_params(
        self,
        psi: torch.Tensor,
        log_mi: torch.Tensor,
        ind_ids_for_etas: Optional[list[str | int]] = None,
    ) -> torch.Tensor:
        """Transform gaussian parameters to physical parameters (thetas)

        Args:
            psi (torch.Tensor): Tensor of individual unconstrained parameter values. Size: (num_chains, nb_patients, nb_PDU)
            ind_ids_for_etas (List[Union[str, int]]): List of individual ids corresponding to the sampled etas

        Returns:
            torch.Tensor: Tensor of individual physical parameter values. Both PDUs and MIs are included. Size: (num_chains, nb_patients, nb_PDU + nb_MI)
        """

        if ind_ids_for_etas is None:
            ind_ids_for_etas = self.patients

        (nb_samples, nb_patients_local, nb_PDU) = psi.shape
        assert nb_patients_local == len(ind_ids_for_etas), nb_PDU == self.nb_PDU

        pdu = torch.zeros_like(psi, device=device)
        # Apply exp transform
        pdu[:, :, self.pdu_transforms["exp"]] = torch.exp(
            psi[:, :, self.pdu_transforms["exp"]]
        )
        # Apply sigmoid transform
        pdu[:, :, self.pdu_transforms["sigmoid"]] = torch.sigmoid(
            psi[:, :, self.pdu_transforms["sigmoid"]]
        )
        # Shift the PDU
        pdu = self.pdu_shift + self.pdu_scale * pdu

        mi = torch.zeros_like(log_mi, device=device)
        # Apply the exp transform
        mi[self.mi_transforms["exp"]] = torch.exp(log_mi[self.mi_transforms["exp"]])
        # Apply sigmoid transform
        mi[self.mi_transforms["sigmoid"]] = torch.sigmoid(
            log_mi[self.mi_transforms["sigmoid"]]
        )
        mi = self.mi_shift + self.mi_scale * mi
        mi = mi.expand(nb_samples, nb_patients_local, -1)

        assert mi.shape == (nb_samples, nb_patients_local, self.nb_MI)
        # Todo: allow for different transformations of gaussian parameters (logit)
        phi = torch.cat((pdu, mi), dim=-1).to(device)

        return phi

    def assemble_individual_parameters(
        self,
        physical_params: torch.Tensor,
        ind_ids_for_etas: Optional[list[str | int]] = None,
    ) -> torch.Tensor:
        """Compute individual patient parameters

        Transforms log(MI) (Model intrinsic), betas: log(mu)s & coeffs for covariates and individual random effects (etas) into individual parameters (theta_i), for each set of etas of the list and corresponding design matrix.
        Assumes log-normal distribution for individual parameters and covariate effects: theta_i[PDU] = mu_pop * exp(eta_i) * exp(covariates_i * cov_coeffs) where eta_i is from N(0, Omega) and theta_i[MI]=MI.

        Args:
            physical_params (Tensor): output from self.gaussian_to_physical_params. Transformed MI and PDU params
            ind_ids_for_etas (List[Union[str, int]]): List of individual ids corresponding to the sampled etas

        Returns:
            torch.Tensor [nb_patients x nb_parameters]: One parameter set for each patient. Dim 0 corresponds to the patients, dim 1 is the parameters
        """
        if ind_ids_for_etas is None:
            ind_ids_for_etas = self.patients

        (nb_samples, nb_patients_local, nb_descriptors) = physical_params.shape

        assert len(ind_ids_for_etas) == nb_patients_local

        if nb_patients_local == self.nb_patients:
            pdk = self.patients_pdk_full
        elif nb_patients_local == 1:
            if self.nb_PDK > 0:
                pdk = self.patients_pdk[ind_ids_for_etas[0]].unsqueeze(0)
            else:
                pdk = torch.empty((nb_patients_local, 0), device=device)
        else:
            raise ValueError(
                f"Unsupported amount of patients requested in `assemble_individual_parameters` {nb_patients_local}."
            )

        # Assemble the thetas by adding the PDKs
        thetas = torch.cat(
            (
                pdk.expand(nb_samples, -1, -1),
                physical_params,
            ),
            dim=-1,
        )

        assert thetas.shape == (nb_samples, nb_patients_local, self.nb_descriptors)
        return thetas

    @torch.compile
    def struc_model_inputs_from_theta(
        self,
        thetas: torch.Tensor,
        ind_ids_for_etas: Optional[list[str | int]] = None,
    ) -> torch.Tensor:
        """Return model inputs for all patients

        Args:
            thetas (torch.Tensor): Parameter values per patient. Theta may contain one parameter per chain
            ind_ids_for_etas (List[Union[str, int]]): List of individual ids corresponding to the sampled etas

        Returns:
            torch.Tensor: the full inputs required to simulate all patients on all time steps
        """
        if ind_ids_for_etas is None:
            ind_ids_for_etas = self.patients

        # If theta was not provided in batch format, add a dummy dimension
        inputs = thetas

        if inputs.dim() < 3:
            inputs = inputs.unsqueeze(0)
        num_chains_local = inputs.shape[0]
        nb_patients_local = len(ind_ids_for_etas)
        assert inputs.shape == (
            num_chains_local,
            nb_patients_local,
            self.nb_descriptors,
        ), f"Wrong shape in theta to structural model inputs: {inputs.shape}"

        if not hasattr(self, "observations_tensors"):
            raise ValueError(
                "Cannot compute patient predictions without an associated observations data frame."
            )

        # Order the columns of theta, and add a repeat dimension to cover time steps
        theta_expanded = (
            inputs[:, :, self.model_input_to_descriptor]
            .unsqueeze(-2)
            .expand((-1, -1, self.nb_global_time_steps, -1))
        )

        # Repeat global time steps as much as there are patients in ind_ids_for_etas
        if nb_patients_local == self.nb_patients:
            time_steps_expanded = self.global_time_steps_expanded
        elif nb_patients_local == 1:
            time_steps_expanded = (
                self.global_time_steps.unsqueeze(0)
                .unsqueeze(-1)
                .repeat((nb_patients_local, 1, 1))
            )
        else:
            raise ValueError(
                "In struc_model_inputs_from_theta, nb_patients_local should either be 1 or the total number of patients."
            )

        full_inputs = torch.cat(
            (
                theta_expanded,
                time_steps_expanded.expand(num_chains_local, -1, -1, -1),
            ),
            dim=-1,
        )

        assert full_inputs.shape == (
            num_chains_local,
            nb_patients_local,
            self.nb_global_time_steps,
            self.nb_descriptors + 1,
        )

        return full_inputs

    def predict_outputs_from_theta(
        self,
        thetas: torch.Tensor,
        ind_ids_for_etas: Optional[list[str | int]] = None,
        model_inputs: Optional[torch.Tensor] = None,
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Return model predictions for all patients

        This function carries the number of chains in the batch dimension (0).

        Args:
            thetas (torch.Tensor): Parameter values per patient. Size (num_chains | 1, nb_patients, nb_time_steps, nb_descriptors + 1)
            ind_ids_for_etas (List[Union[str, int]]): List of individual ids corresponding to the sampled etas

        Returns:
            list[torch.Tensor]: a tensor of predictions for each patient
        """
        if ind_ids_for_etas is None:
            ind_ids_for_etas = self.patients

        nb_patients_local = len(ind_ids_for_etas)

        if model_inputs is None:
            model_inputs = self.struc_model_inputs_from_theta(thetas)

        if nb_patients_local == self.nb_patients:
            prediction_index = self.prediction_index
            chunk_sizes = self.chunk_sizes
        elif nb_patients_local == 1:
            ind = ind_ids_for_etas[0]
            prediction_index = (
                torch.zeros_like(
                    self.observations_tensors[ind]["p_index_repeated"]
                ),  # only one patient, index 0
                self.observations_tensors[ind]["time_step_indices"],
                self.observations_tensors[ind]["tasks_indices"],
            )
            chunk_sizes = [len(prediction_index[0])]
        else:
            raise ValueError(
                "In predict_outputs_from_theta, nb_patients_local should either be 1 or the total number of patients."
            )

        # Shape: (nb_chains | 1, nb_patients, nb_global_time_steps, nb_descriptors + 1)
        pred_mean, pred_var = self.structural_model.simulate(
            model_inputs,
            prediction_index,
            chunk_sizes,
        )

        return pred_mean, pred_var

    def add_residual_error(self, outputs: torch.Tensor) -> torch.Tensor:
        res_var = self.residual_var.index_select(0, self.full_output_indices)
        noise = torch.distributions.Normal(
            torch.zeros(self.full_output_indices.shape[0], device=device), res_var
        ).sample()
        if self.error_model_type == "additive":
            new_out = outputs + noise
        elif self.error_model_type == "proportional":
            new_out = outputs * noise
        else:
            raise ValueError(f"Non-implemented error model {self.error_model_type}")
        return new_out

    def outputs_to_df(self, outputs: torch.Tensor) -> pd.DataFrame:
        """Transform the NLME model outputs to a data frame in order to compare with observed data

        Args:
            outputs (torch.Tensor): Outputs from `self.predict_outputs_from_theta`

        Returns:
            pd.DataFrame: A data frame containing the following columns
            - `id`
            - `output_name`
            - `protocol_arm`
            - `time`
            - `predicted_value`
        """
        assert (
            outputs.shape[0] == 1
        ), "Batched outputs cannot be converted to a dataframe."
        outputs_list = outputs.squeeze(0).cpu().split(self.chunk_sizes)
        df_list = []
        for ind_idx, ind in enumerate(self.patients):
            time_steps = self.observations_tensors[ind]["time_steps"]
            task_list = self.observations_tensors[ind]["tasks_indices"]
            temp_df = pd.DataFrame(
                {
                    "time": time_steps.cpu().numpy(),
                    "id": ind,
                    "task_index": task_list.cpu(),
                    "predicted_value": outputs_list[ind_idx].numpy(),
                }
            )
            temp_df["output_name"] = temp_df["task_index"].apply(
                lambda t: self.outputs_names[
                    self.structural_model.task_idx_to_output_idx[t]
                ]
            )
            temp_df["protocol_arm"] = temp_df["task_index"].apply(
                lambda t: self.structural_model.task_idx_to_protocol[t]
            )
            df_list.append(temp_df)
        out_df = pd.concat(df_list)
        out_df = out_df.drop(columns=["task_index"])
        return out_df

    def _log_prior_etas(self, etas: torch.Tensor) -> torch.Tensor:
        """Compute log-prior of random effect samples (etas)

        Args:
            etas (torch.Tensor): Individual samples, assuming eta_i ~ N(0, Omega)

        Returns:
            torch.Tensor [nb_eta_i x nb_PDU]: Values of log-prior, computed according to:

            P(eta) = (1/sqrt((2pi)^k * |Omega|)) * exp(-0.5 * eta.T * omega.inv * eta)
            log P(eta) = -0.5 * (k * log(2pi) + log|Omega| + eta.T * omega.inv * eta)

        """

        log_priors: torch.Tensor = self.eta_distribution.log_prob(etas).to(device)
        return log_priors

    def log_posterior_etas(
        self,
        etas: torch.Tensor,
    ) -> tuple[
        torch.Tensor,
        torch.Tensor,
        torch.Tensor,
    ]:
        """Compute the log-posterior of a list of random effects

        Args:
            etas (torch.Tensor): Random effects samples

        Returns:
            tuple[torch.Tensor, list[torch.Tensor], DataFrame]:
            - log-posterior likelihood of etas
            - gaussian parameters corresponding to the etas
            - list of simulated values for each patient on each chain

        """
        nb_chains_local = etas.shape[0]
        assert etas.shape == (nb_chains_local, self.nb_patients, self.nb_PDU)
        if not hasattr(self, "observations_tensors"):
            raise ValueError(
                "Cannot compute log-posterior without an associated observations data frame."
            )
        gaussian_params_chains = self.etas_to_gaussian_params(etas)
        physical_params_chains = self.gaussian_to_physical_params(
            gaussian_params_chains, self.log_MI
        )
        # Get individual parameters in a tensor
        theta_chains = self.assemble_individual_parameters(physical_params_chains)

        # Run the structural model
        full_pred, full_var = self.predict_outputs_from_theta(theta_chains)
        # var_list = full_var.split(self.chunk_sizes)
        # Validate the variance magnitude
        # warnings = self.variance_level_check(var_list, self.pred_var_threshold)
        # flagged_patients = [individual_params[i, :].tolist() for i in warnings]

        # calculate log-prior of the random samples
        log_priors: torch.Tensor = self._log_prior_etas(etas)
        assert log_priors.shape == (nb_chains_local, self.nb_patients)

        # Compute the log likelihood of observations
        log_likelihood_observations = cast(
            torch.Tensor, self.log_likelihood_observation(full_pred)
        )
        assert log_likelihood_observations.shape == (nb_chains_local, self.nb_patients)

        log_posterior = log_likelihood_observations + log_priors
        assert log_posterior.shape == (nb_chains_local, self.nb_patients)
        return (
            log_posterior,
            gaussian_params_chains,
            full_pred,
        )

    @torch.compile
    def calculate_residuals(
        self,
        observed_data: torch.Tensor,
        predictions: torch.Tensor,
    ) -> torch.Tensor:
        """Calculates residuals based on the error model

        Args:
            observed_data: Tensor of observations
            predictions: Tensor of predictions

        Returns:
            torch.Tensor: a tensor of residual values
        """
        assert (
            predictions.shape == observed_data.shape
        ), f"Non-matching shapes in `calculate_residuals`: {predictions.shape=}, {observed_data.shape=}"

        if self.error_model_type == "additive":
            return observed_data - predictions
        elif self.error_model_type == "proportional":
            return (observed_data - predictions) / predictions
        else:
            raise ValueError("Unsupported error model type.")

    @torch.compile
    def sum_sq_residuals_chains(self, prediction: torch.Tensor) -> torch.Tensor:
        """Compute the sum of squared residuals for current predictions on all chains

        Args:
            prediction (torch.Tensor): Predictions for all patients and all chains, Size (num_chains, nb_patients, nb_tot_obs)

        Returns:
            torch.Tensor: The sum of squared residuals for each output. Size (nb_outputs)
        """
        assert prediction.shape == (self.num_chains, self.nb_tot_obs)
        sq_residuals = torch.square(
            self.calculate_residuals(
                self.full_obs_data.expand(self.num_chains, -1), prediction
            )
        )
        sum_residuals_per_chain = torch.zeros(
            self.num_chains, self.nb_outputs, device=device
        )
        sum_residuals_per_chain.scatter_add_(
            1, self.full_output_indices.expand(self.num_chains, -1), sq_residuals
        )
        sum_residuals = sum_residuals_per_chain.mean(dim=0)
        return sum_residuals

    @torch.compile
    def compute_error_variance(
        self, predictions: torch.Tensor, output_indices: Optional[torch.Tensor] = None
    ) -> torch.Tensor:

        if output_indices is None:
            output_indices = self.full_output_indices

        num_chains_local = predictions.shape[0]

        res_error_var = self.residual_var.expand(num_chains_local, -1).index_select(
            1, output_indices
        )

        # Log-likelihood of normal distribution
        if self.error_model_type == "additive":
            variance = res_error_var
        elif self.error_model_type == "proportional":
            variance = res_error_var * torch.square(predictions)
        else:
            raise ValueError("Non supported error type.")
        return variance

    def compute_iwres(
        self,
    ) -> dict:
        """Compute Individual Weighted RESiduals, following the formula :

        IWRES_(ij) = ( y_ij - f(t_ij, psi_i) ) / g(t_ij, psi_i)

        Returns:
            dict: IWRES with patientId as key, with IWRES and timesteps for each patient
        """

        # Avoid computing if already computed residuals
        if hasattr(self, "iwres"):
            return self.iwres

        # Gather observations tensor
        observed_tensor = self.full_obs_data.view(1, -1)

        # Gather predictions tensor
        ebe_physical = self.compute_ebe()
        (nb_samples, nb_patients_local, nb_descriptors) = ebe_physical.shape

        # Assemble the thetas by adding the PDKs
        theta = torch.cat(
            (
                self.patients_pdk_full.expand(nb_samples, -1, -1),
                ebe_physical,
            ),
            dim=-1,
        )
        simulated_tensor, _ = self.predict_outputs_from_theta(theta)

        # Compute residuals and variance
        residuals = self.calculate_residuals(observed_tensor, simulated_tensor)
        variance = self.compute_error_variance(simulated_tensor)

        iwres = residuals / torch.sqrt(variance)

        iwres_list = torch.split(iwres, self.chunk_sizes, dim=1)
        iwres_results = {}

        # Separate IWRES per patient in a dict
        for i, patient_id in enumerate(self.patients):
            iwres_patient = iwres_list[i]
            time_steps_patient = self.observations_tensors[patient_id]["time_steps"]
            iwres_results.update(
                {
                    patient_id: {
                        "iwres": iwres_patient.squeeze().cpu().numpy(),
                        "time": time_steps_patient.cpu().numpy(),
                    }
                }
            )

        self.iwres = iwres_results
        return iwres_results

    def compute_pwres(self, num_samples: int) -> dict:
        """Compute Population Weighted RESiduals, following the formula :

        PWRES_i = V_i^(-1/2) (y_i - E(f(t_ij, psi_i))

        Returns:
            dict: PWRES with patientId as key, with PWRES and timesteps for each patient
        """

        # Avoid computing if already computed residuals
        if hasattr(self, "pwres"):
            return self.pwres

        # Sample new etas, in order to approximate mean E(y_i) and variance V_i
        mc_etas = self.sample_etas(num_samples, self.nb_patients)

        # Compute gaussian parameters
        mc_gaussian = self.etas_to_gaussian_params(mc_etas)

        # Compute physical parameters
        mc_physical = self.gaussian_to_physical_params(mc_gaussian, self.log_MI)

        # Assemble individual parameters
        mc_thetas = self.assemble_individual_parameters(mc_physical)

        # Simulate outputs
        simulated_tensor, _ = self.predict_outputs_from_theta(mc_thetas)

        # Compute PWRES per patient

        # list_sim_per_patient shape: n_patients items of shape nb_samples * n_obs_patient
        list_sim_per_patient = torch.split(simulated_tensor, self.chunk_sizes, dim=1)

        # Initialize a dictionnary gathering PWRES and timesteps for each patient, to plot later
        pwres_results = {}

        for i, patient_id in enumerate(self.patients):

            # patient_data shape: nb_samples * n_obs_patient
            patient_data = list_sim_per_patient[i]

            # mean_patient shape: nb_samples * n_obs_patient -> n_obs_patient
            mean_patient = patient_data.mean(dim=0)

            # variance_patient shape: n_obs_patient * n_obs_patient
            variance_patient = torch.cov(patient_data.T)

            # obs_patient shape: n_obs_patient
            obs_patient = self.observations_tensors[patient_id]["observations"]
            time_steps_patient = self.observations_tensors[patient_id]["time_steps"]

            # Transform residual into a column
            residual = (obs_patient - mean_patient).unsqueeze(-1)

            # Compute V^-1/2 with Cholesky factorization, adding a jitter for stability purposes
            jitter = torch.eye(variance_patient.size(0)) * 1e-6
            L = torch.linalg.cholesky(variance_patient + jitter)

            # Compute patient PWRES and add them to dictionnary
            pwres_patient = L @ residual
            pwres_results.update(
                {
                    patient_id: {
                        "pwres": pwres_patient.squeeze(-1).cpu().numpy(),
                        "time": time_steps_patient.cpu().numpy(),
                    }
                }
            )

        self.pwres = pwres_results
        return pwres_results

    def compute_npde(self, num_samples: int = 100) -> dict:

        # Avoid computing if already computed residuals
        if hasattr(self, "npde"):
            return self.npde

        # Sample new etas
        mc_etas = self.sample_etas(num_samples, self.nb_patients)

        # Compute gaussian parameters
        mc_gaussian = self.etas_to_gaussian_params(mc_etas)

        # Compute physical parameters
        mc_physical = self.gaussian_to_physical_params(mc_gaussian, self.log_MI)

        # Assemble individual parameters
        mc_thetas = self.assemble_individual_parameters(mc_physical)

        # Simulate outputs
        simulated_tensor, _ = self.predict_outputs_from_theta(mc_thetas)

        # Expand observation tensor to match simulated tensor
        observed_tensor = self.full_obs_data.expand(num_samples, -1)

        # Compute indicator function in NPDE formula
        mc_F = simulated_tensor <= observed_tensor
        mc_F = mc_F.to(torch.float)

        # Average on MC samples, avoiding 0 and 1 values
        mean_F = mc_F.mean(dim=0)
        eps = 1.0 / simulated_tensor.shape[0]
        mean_F_clamped = torch.clamp(mean_F, min=eps, max=1.0 - eps)

        # Apply normal inverse CDF to compare NPDE with N(0,1)
        normal_dist = torch.distributions.Normal(0, 1)
        npde = normal_dist.icdf(mean_F_clamped)
        npde_per_patient = torch.split(npde, self.chunk_sizes, dim=0)

        npde_results = {}

        for i, patient_id in enumerate(self.patients):

            npde_patient = npde_per_patient[i]
            time_steps_patient = self.observations_tensors[patient_id]["time_steps"]
            npde_results.update(
                {
                    patient_id: {
                        "npde": npde_patient.squeeze(-1).cpu().numpy(),
                        "time": time_steps_patient.cpu().numpy(),
                    }
                }
            )

        self.npde = npde_results
        return npde_results

    def log_likelihood_observation(
        self,
        predictions: torch.Tensor,
        ind_ids_for_etas: Optional[list[str | int]] = None,
    ) -> torch.Tensor:
        """Compute the log-likelihood of observations across all chains

        Args:
            predictions (torch.Tensor): Tensor of predictions. Size (num_chains, nb_total_obs, 1)

        Returns:
            torch.Tensor: Log likelihood per patient and per chain. Size (num_chains, nb_patient)
        """
        num_chains_local = predictions.shape[0]

        if ind_ids_for_etas is None:
            ind_ids_for_etas = self.patients

        nb_patients_local = len(ind_ids_for_etas)
        if nb_patients_local == 1:
            patient_id = ind_ids_for_etas[0]
            observations = self.observations_tensors[patient_id]["observations"]
            output_indices = self.observations_tensors[patient_id]["outputs_indices"]
        elif nb_patients_local == self.nb_patients:
            observations = self.full_obs_data
            output_indices = self.full_output_indices
        else:
            raise ValueError(
                f"Unsupported amount of patients requested in `log_likelihood_observation` {nb_patients_local}."
            )

        residuals: torch.Tensor = self.calculate_residuals(
            observations.expand(num_chains_local, -1), predictions
        )

        res_error_var = self.residual_var.expand(num_chains_local, -1).index_select(
            1, output_indices
        )

        assert residuals.shape == res_error_var.shape

        # Log-likelihood of normal distribution
        variance = self.compute_error_variance(predictions, output_indices)

        log_lik_full = -0.5 * (
            torch.log(2 * torch.pi * variance) + (residuals**2 / variance)
        )

        # If only one patient, return likelihood as a single item tensor
        if nb_patients_local == 1:
            log_lik_patient = log_lik_full.sum()
            return log_lik_patient

        # If several patients, create likelihood tensor
        else:
            log_lik_per_patient = torch.zeros(
                (num_chains_local, self.nb_patients), device=device
            )
            log_lik_per_patient.scatter_add_(
                1,
                self.observation_to_patient_index.expand(num_chains_local, -1),
                log_lik_full,
            )
            return log_lik_per_patient

    def mh_step(
        self,
        current_etas: torch.Tensor,
        current_log_prob: torch.Tensor,
        current_pred: torch.Tensor,
        current_gaussian_params: torch.Tensor,
        step_size: float,
        learning_rate: float,
        target_acceptance_rate: float = 0.234,
        verbose: bool = False,
    ) -> tuple[
        torch.Tensor,
        torch.Tensor,
        torch.Tensor,
        torch.Tensor,
        torch.Tensor,
        float,
    ]:
        """Perform one step of a Metropolis-Hastings transition kernel

        Args:
            current_etas (torch.Tensor): values of the individual random effects for all patients, on all chains. Size (num_chains, nb_patients, nb_PDU)
            current_log_prob (torch.Tensor): log posterior likelihood of current random effects. Size (num_chains, nb_patients)
            current_pred (torch.Tensor): associated model predictions with current random effects. Size (num_chains, nb_obs_total)
            step_size (torch.Tensor): current value of MH step size,
            learning_rate (float): current learning rate (defined by the optimization algorithm)
            target_acceptance_rate (float, optional): Target for the MCMC acceptance rate. Defaults to 0.234 [1].

            [1] Sherlock C. Optimal Scaling of the Random Walk Metropolis: General Criteria for the 0.234 Acceptance Rule. Journal of Applied Probability. 2013;50(1):1-15. doi:10.1239/jap/1363784420

        Returns:
            tuple[torch.Tensor, torch.Tensor, dict[int | str, torch.Tensor], torch.Tensor, float]:
            - updated individual random effects
            - updated log posterior likelihood
            - updated predictions, for each patient of the observation data set
            - updated thetas
            - updated values of log PDUs
            - updated step size
            - a dict of warnings for all patients with predictive variance above threshold
        """
        # Propose new etas
        proposal_noise = (
            torch.randn_like(current_etas, device=device) @ self.omega_pop_lower_chol
        )
        proposal_etas = current_etas + step_size * proposal_noise
        # Compute their log posterior likelihood
        (
            proposal_log_prob,
            proposal_gaussian_params,
            proposal_pred,
        ) = self.log_posterior_etas(proposal_etas)

        assert proposal_log_prob.shape == current_log_prob.shape

        # Define acceptance masks
        deltas: torch.Tensor = proposal_log_prob - current_log_prob
        log_u: torch.Tensor = torch.log(torch.rand_like(deltas, device=device))
        accept_mask: torch.Tensor = log_u < deltas
        accept_mask_parameters = accept_mask.unsqueeze(-1).expand(
            -1, -1, current_etas.shape[-1]
        )
        accept_mask_predictions = accept_mask.index_select(
            1, self.observation_to_patient_index
        )
        # Accept the different variables using the masks
        new_etas = torch.where(accept_mask_parameters, proposal_etas, current_etas).to(
            device
        )
        new_gaussian_params = torch.where(
            accept_mask_parameters, proposal_gaussian_params, current_gaussian_params
        ).to(device)
        new_log_prob = torch.where(accept_mask, proposal_log_prob, current_log_prob).to(
            device
        )
        # Compute the complete likelihood by averaging over chains and summing over patients
        new_complete_likelihood = -2 * new_log_prob.mean(dim=0).sum(dim=0)
        new_pred = torch.where(accept_mask_predictions, proposal_pred, current_pred).to(
            device
        )
        new_acceptance_rate: float = accept_mask.cpu().float().mean().mean().item()
        if verbose:
            print(f"  Acceptance rate: {new_acceptance_rate:.2f}")
        new_step_size: float = step_size * np.exp(
            learning_rate * (new_acceptance_rate - target_acceptance_rate)
        )
        return (
            new_etas,
            new_log_prob,
            new_complete_likelihood,
            new_pred,
            new_gaussian_params,
            new_step_size,
        )

    def map_estimates_descriptors(self) -> pd.DataFrame:
        ebe_physical = self.compute_ebe()

        (nb_samples, nb_patients, nb_descriptors) = ebe_physical.shape

        # Assemble the thetas by adding the PDKs
        theta = torch.cat(
            (
                self.patients_pdk_full.expand(nb_samples, -1, -1),
                ebe_physical,
            ),
            dim=-1,
        )

        if theta is None:
            raise ValueError("No estimation available yet. Run the algorithm first.")

        theta = theta.squeeze()

        assert theta.shape == (
            self.nb_patients,
            self.nb_descriptors,
        ), f"Unexpected dimensions of MAP estimates: {theta.shape}"

        map_per_patient = pd.DataFrame(
            data=theta.cpu().numpy(), columns=self.descriptors
        )
        id_column = self.patients
        map_per_patient["id"] = id_column

        return map_per_patient

    def map_estimates_predictions(self) -> pd.DataFrame:
        ebe_physical = self.compute_ebe()

        (nb_samples, nb_patients, nb_descriptors) = ebe_physical.shape

        # Assemble the thetas by adding the PDKs
        theta = torch.cat(
            (
                self.patients_pdk_full.expand(nb_samples, -1, -1),
                ebe_physical,
            ),
            dim=-1,
        )

        if theta is None:
            raise ValueError(
                "No estimation available yet. Run the optimization algorithm first."
            )
        simulated_tensor, _ = self.predict_outputs_from_theta(theta)
        simulated_df = self.outputs_to_df(simulated_tensor)
        return simulated_df

    def population_predictions(self) -> pd.DataFrame:
        """
        Returns simulated dataframe for null etas, i.e. without individual random effects but keeping the covariables effects.
        """

        # Sample new etas, in order to approximate mean E(y_i) and variance V_i
        null_etas = torch.zeros([1, self.nb_patients, self.nb_PDU])

        # Compute gaussian parameters
        physical_params = self.etas_to_gaussian_params(null_etas)

        # Compute physical parameters
        physical_params = self.gaussian_to_physical_params(physical_params, self.log_MI)

        # Assemble individual parameters
        thetas = self.assemble_individual_parameters(physical_params)

        # Simulate outputs
        simulated_tensor, _ = self.predict_outputs_from_theta(thetas)

        simulated_df = self.outputs_to_df(simulated_tensor)
        return simulated_df

    def variance_level_check(
        self, var_list: tuple[torch.Tensor, ...], threshold: float
    ) -> list:
        warnings = []
        for i in range(self.nb_patients):
            var, _ = var_list[i].cpu().max(0)
            if var > threshold:
                warnings.append(i)
        return warnings

    def sample_conditional_distribution(
        self,
        nb_samples: int = 100,
        nb_burn_in: int = 50,
        override: bool = False,
    ):
        """
        Returns: cond_dist_samples: dim(nb_samples, nb_patients, nb_PDU)
        """

        if (
            hasattr(self, "cond_dist_samples")
            and nb_samples == self.cond_dist_samples.shape[0]
            and (not override)
        ):
            return self.cond_dist_samples

        if smoke_test:
            nb_samples = 2
            nb_burn_in = 1

        init_etas = self.sample_etas(1, self.nb_patients)
        (current_log_prob, current_gaussian_params, current_pred) = (
            self.log_posterior_etas(init_etas)
        )
        current_etas = init_etas
        sample_list = []
        print(f"Sampling conditional distribution on {nb_samples} samples:")
        for i in tqdm(range(nb_burn_in + nb_samples)):
            (
                current_etas,
                current_log_prob,
                _,
                current_pred,
                current_gaussian_params,
                _,
            ) = self.mh_step(
                current_etas,
                current_log_prob,
                current_pred,
                current_gaussian_params,
                step_size=0.1,
                learning_rate=0.0,
            )

            if i >= nb_burn_in:
                sample_list.append(current_etas)

        sample_tensor = torch.stack(sample_list)
        etas_samples = torch.squeeze(sample_tensor, 1)
        self.cond_dist_samples = etas_samples
        return self.cond_dist_samples

    def compute_ebe(self, max_iter: int = 5000, override: bool = False):
        """
        Returns: ebe_estimates: dim(nb_patients, nb_PDU)
        """

        if hasattr(self, "ebe_estimates") and (not override):
            return self.ebe_estimates

        samples = self.sample_conditional_distribution(100)

        if smoke_test:
            max_iter = 1

        # Taking conditional distribution samples means as a starting point for optimization
        init_samples = samples.mean(dim=0)

        nb_PDU = self.nb_PDU
        ebe_etas = torch.zeros((self.nb_patients, self.nb_PDU))
        print("Computing EBEs for each patient:")
        for i in tqdm(range(self.nb_patients)):

            def objective_function(eta_array):
                eta_tensor = torch.from_numpy(eta_array).float().view(1, 1, nb_PDU)
                with torch.no_grad():
                    log_post = self.log_posterior_etas_single(eta_tensor, i)
                return -log_post

            x0 = init_samples[i].numpy()

            res = minimize(
                objective_function,
                x0,
                method="Nelder-Mead",
                tol=1e-4,
                options={"maxiter": max_iter},
            )
            ebe_etas[i] = torch.from_numpy(res.x)

        ebe_etas = ebe_etas.expand(1, -1, -1)
        ebe_gaussian = self.etas_to_gaussian_params(ebe_etas)
        ebe_physical = self.gaussian_to_physical_params(ebe_gaussian, self.log_MI)
        self.ebe_estimates = ebe_physical
        return ebe_physical

    def log_posterior_etas_single(self, etas: torch.Tensor, patient_ind: int) -> float:
        """
        Args:
            etas: random effects sample for one patient, dim (1, 1, nb_PDU)
            patient_ind: index of a patient in self.patients

        Returns:
            log_posterior: value of log-posterior distribution for given etas and patient
        """

        assert etas.shape == (1, 1, self.nb_PDU)

        patient_id = self.patients[patient_ind]

        if not hasattr(self, "observations_tensors"):
            raise ValueError(
                "Cannot compute log-posterior without an associated observations data frame."
            )

        # Proceed from etas to model inputs
        gaussian_params = self.etas_to_gaussian_params(etas, [patient_id])
        physical_params = self.gaussian_to_physical_params(
            gaussian_params, self.log_MI, [patient_id]
        )
        theta = self.assemble_individual_parameters(physical_params, [patient_id])
        full_inputs = self.struc_model_inputs_from_theta(theta, [patient_id])

        predictions, _ = self.predict_outputs_from_theta(
            theta, [patient_id], full_inputs
        )

        log_priors = self._log_prior_etas(etas)
        log_lik_patient = self.log_likelihood_observation(
            predictions, [patient_id]
        ).item()

        log_posterior = cast(float, log_lik_patient + log_priors)

        return log_posterior



import math
import torch
import numpy as np
import pandas as pd
from functools import reduce

from ..utils import device


class TrainingDataSet:
    def __init__(
        self,
        training_df: pd.DataFrame,
        descriptors: list[str],
        training_proportion: float = 0.7,
        log_lower_limit: float = 1e-10,
        log_inputs: list[str] = [],
        log_outputs: list[str] = [],
        data_already_normalized: bool = False,
    ):
        """Instantiate a TrainingDataSet container

        The data container is used to process training data, normalize inputs and outputs, and provide utilitary methods to transform inputs and outputs for a PyTorch model. In particular, it transforms output names and protocol arms into individual tasks by cartesian product, and it provides mappings between tasks and protcol/output.

        Args:
            training_df (pd.DataFrame): the training data. Should contain the columns [`id`, `output_name`, `protocol_arm`, *descriptors, `value`]
            descriptors (list[str]): the names of the columns of `training_df` which correspond to descriptors on which to train the model
            training_proportion (float, optional): Proportion of patients to be used as training vs. validation. Defaults to 0.7.
            log_lower_limit(float): epsilon value that is added to all rescaled value to avoid numerical errors when log-scaling variables
            log_inputs (list[str]): the list of parameter inputs which should be rescaled to log when fed to the GP. Avoid adding time here, or any parameter that takes 0 as a value.
            log_outputs (list[str]): list of model outptus which should be rescaled to log
            data_already_normalized(bool): set to True if the data set is preprocessed and no normalization / scaling is required
        """

        # Process the supplied data set
        self.full_df_raw = training_df

        declared_columns = self.full_df_raw.columns.to_list()
        # Input validation
        if not ("id" in declared_columns):
            raise ValueError("Training data should contain an `id` column.")
        if not ("output_name" in declared_columns):
            raise ValueError("Training data should contain an `output_name` column.")
        if not ("value" in declared_columns):
            raise ValueError("Training data should contain a `value` column.")
        if not set(descriptors) <= set(declared_columns):
            raise ValueError(
                f"The provided inputs are not declared in the data set: {descriptors}."
            )

        self.parameter_names = descriptors
        self.training_proportion = training_proportion
        self.nb_parameters = len(self.parameter_names)
        self.data_already_normalized = data_already_normalized
        if not ("protocol_arm" in declared_columns):
            self.full_df_raw["protocol_arm"] = "identity"
        self.protocol_arms = self.full_df_raw["protocol_arm"].unique().tolist()
        self.nb_protocol_arms = len(self.protocol_arms)
        self.output_names = self.full_df_raw["output_name"].unique().tolist()
        self.nb_outputs = len(self.output_names)
        self.log_lower_limit = log_lower_limit
        self.log_inputs = log_inputs
        self.log_inputs_indices = [
            self.parameter_names.index(p) for p in self.log_inputs
        ]
        self.log_outputs = log_outputs

        # Ensure input df has a consistent shape (and remove potential extra columns)
        self.full_df_raw = self.full_df_raw[
            ["id"] + self.parameter_names + ["output_name", "protocol_arm", "value"]
        ]

        # Gather the list of patients in the training data
        self.patients = self.full_df_raw["id"].unique()
        self.nb_patients = self.patients.shape[0]

        print(
            f"Successfully loaded a training data set with {self.nb_patients} patients. The following outputs are available:\n{self.output_names}\n and the following protocol arms:\n{self.protocol_arms}"
        )
        # Construct the list of tasks, mapping from output name and protocol arm to task number
        self.tasks: list[str] = [
            output + "_" + protocol
            for protocol in self.protocol_arms
            for output in self.output_names
        ]
        self.nb_tasks = len(self.tasks)
        # Map tasks to output names
        self.task_to_output = {
            output_name + "_" + protocol_arm: output_name
            for output_name in self.output_names
            for protocol_arm in self.protocol_arms
        }
        # Map task index to output index
        self.task_idx_to_output_idx = {
            self.tasks.index(k): self.output_names.index(v)
            for k, v in self.task_to_output.items()
        }
        # Map task to protocol arm
        self.task_to_protocol = {
            output_name + "_" + protocol_arm: protocol_arm
            for output_name in self.output_names
            for protocol_arm in self.protocol_arms
        }
        # Map task index to protocol arm
        self.task_idx_to_protocol = {
            self.tasks.index(k): v for k, v in self.task_to_protocol.items()
        }
        # list tasks that should be rescaled to log
        self.log_tasks = [
            task for task in self.tasks if self.task_to_output[task] in self.log_outputs
        ]
        self.log_tasks_indices = [self.tasks.index(task) for task in self.log_tasks]

        ## Data processing
        # Pivot the data to the correct shape for GP training
        self.full_df_reshaped = self.pivot_input_data(self.full_df_raw)

        # Normalize the inputs and the outputs (only if required)
        if self.data_already_normalized == True:
            self.normalized_df = self.full_df_reshaped
        else:
            self.full_df_reshaped[self.log_inputs + self.log_tasks] = (
                self.full_df_reshaped[self.log_inputs + self.log_tasks].apply(
                    lambda val: np.log(np.maximum(val, self.log_lower_limit))
                )
            )

            self.normalized_df, mean, std = self.normalize_data(
                self.full_df_reshaped, ["id"]
            )
            self.normalizing_input_mean, self.normalizing_input_std = (
                torch.as_tensor(mean.loc[self.parameter_names].values, device=device),
                torch.as_tensor(std.loc[self.parameter_names].values, device=device),
            )
            self.normalizing_output_mean, self.normalizing_output_std = (
                torch.as_tensor(mean.loc[self.tasks].values, device=device),
                torch.as_tensor(std.loc[self.tasks].values, device=device),
            )

        # Compute the number of patients for training
        self.nb_patients_training = math.floor(
            self.training_proportion * self.nb_patients
        )
        self.nb_patients_validation = self.nb_patients - self.nb_patients_training

        if self.training_proportion != 1:  # non-empty validation data set
            if self.nb_patients_training == self.nb_patients:
                raise ValueError(
                    "Training proportion too high for the number of sets of parameters: all would be used for training. Set training_proportion as 1 if that is your intention."
                )

            # Randomly mixing up patients
            mixed_patients = np.random.permutation(self.patients)

            self.training_patients = mixed_patients[: self.nb_patients_training]
            self.validation_patients = mixed_patients[self.nb_patients_training :]

            self.training_df_normalized: pd.DataFrame = self.normalized_df.loc[
                self.normalized_df["id"].isin(self.training_patients)
            ]
            self.validation_df_normalized: pd.DataFrame = self.normalized_df.loc[
                self.normalized_df["id"].isin(self.validation_patients)
            ]
            self.X_validation = torch.as_tensor(
                self.validation_df_normalized[self.parameter_names].values,
                device=device,
            )
            self.Y_validation = torch.as_tensor(
                self.validation_df_normalized[self.tasks].values, device=device
            )

        else:  # no validation data set provided
            self.training_patients = self.patients
            self.training_df_normalized = self.normalized_df
            self.validation_df = None
            self.X_validation = None
            self.Y_validation = None

        self.X_training: torch.Tensor = torch.as_tensor(
            self.training_df_normalized[self.parameter_names].values, device=device
        )
        self.Y_training: torch.Tensor = torch.as_tensor(
            self.training_df_normalized[self.tasks].values, device=device
        )

    def pivot_input_data(self, data_in: pd.DataFrame) -> pd.DataFrame:
        """Pivot and reorder columns from a data frame to feed to the model

        This method is used at initialization on the training data frame), and when plotting the model performance against existing data.

        Args:
            data_in (pd.DataFrame): Input data frame, containing the following columns
            - `id`: patient id
            - one column per descriptor, the same descriptors as self.parameter_names should be present
            - `output_name`: the name of the output
            - `protocol_arm`: the name of the protocol arm
            - `value`: the observed value

        Returns:
            pd.DataFrame: A validated and pivotted dataframe with one column per task (`outputName_protocolArm`), and one row per observation
        """

        # util function to rename columns as `output_protocol`
        def join_if_two(tup: str) -> str:
            if tup[0] == "":
                return tup[1]
            elif tup[1] == "":
                return tup[0]
            else:
                return "_".join(tup)

        # Pivot the data set
        reshaped_df = data_in.pivot(
            index=["id"] + self.parameter_names,
            columns=["output_name", "protocol_arm"],
            values="value",
        ).reset_index()
        nested_column_names = reshaped_df.columns.to_list()
        flat_column_names = list(map(join_if_two, nested_column_names))
        reshaped_df.columns = flat_column_names

        assert set(reshaped_df.columns) == set(
            ["id"] + self.parameter_names + self.tasks
        ), "Incomplete training data set provided."

        return reshaped_df

    def normalize_data(
        self, data_in: pd.DataFrame, ignore: list[str]
    ) -> tuple[pd.DataFrame, pd.Series, pd.Series]:
        """Normalize a data frame with respect to its mean and std, ignoring certain columns."""
        selected_columns = data_in.columns.difference(ignore)
        norm_data = data_in
        mean = data_in[selected_columns].mean()
        std = data_in[selected_columns].std()
        norm_data[selected_columns] = (norm_data[selected_columns] - mean) / std
        return norm_data, mean, std

    @torch.compile
    def unnormalize_output_wide(self, data: torch.Tensor) -> torch.Tensor:
        """Unnormalize wide outputs (all tasks included) from the model."""
        unnormalized = data * self.normalizing_output_std + self.normalizing_output_mean
        unnormalized[:, self.log_tasks_indices] = torch.exp(
            unnormalized[:, self.log_tasks_indices]
        )

        return unnormalized

    @torch.compile
    def unnormalize_output_long(
        self, data: torch.Tensor, task_indices: torch.LongTensor
    ) -> torch.Tensor:
        """Unnormalize long outputs (one row per task) from the model."""
        rescaled_data = data
        for task_idx in range(self.nb_tasks):
            log_task = self.tasks[task_idx] in self.log_tasks
            mask = torch.tensor(task_indices == task_idx, device=device).bool()
            rescaled_data[mask] = (
                rescaled_data[mask] * self.normalizing_output_std[task_idx]
                + self.normalizing_output_mean[task_idx]
            )
            if log_task:
                rescaled_data[mask] = torch.exp(rescaled_data[mask])
        return rescaled_data

    @torch.compile
    def normalize_inputs_tensor(self, inputs: torch.Tensor) -> torch.Tensor:
        """Normalize new inputs provided to the model as a tensor. The columns of the input tensor should be the same as [self.descriptors]"""
        X = inputs.to(device)
        X[:, self.log_inputs_indices] = torch.log(X[:, self.log_inputs_indices])
        mean = self.normalizing_input_mean
        std = self.normalizing_input_std
        norm_X = (X - mean) / std

        return norm_X

    def pivot_outputs_longer(
        self, comparison_df: pd.DataFrame, Y: torch.Tensor, name: str
    ) -> pd.DataFrame:
        """Given wide outputs from a model and a comparison data frame (wide format), add the patient descriptors and reshape to a long format, with a `protocol_arm` and an `output_name` column."""
        # Assuming Y is a wide output from the model, its columns are self.tasks
        base_df = pd.DataFrame(
            data=Y.cpu().detach().float().numpy(),
            columns=self.tasks,
        )
        # The rows are assumed to correspond to the rows of the comparison data frame
        base_df[["id"] + self.parameter_names] = comparison_df[
            ["id"] + self.parameter_names
        ]
        # Pivot the data frame to a long format, separating the task names into protocol arm and output name
        long_df = (
            pd.wide_to_long(
                df=base_df,
                stubnames=self.output_names,
                i=["id"] + self.parameter_names,
                j="protocol_arm",
                sep="_",
                suffix=".*",
            )
            .reset_index()
            .melt(
                id_vars=["id"] + self.parameter_names + ["protocol_arm"],
                value_vars=self.output_names,
                var_name="output_name",
                value_name=name,
            )
        )
        return long_df

    def get_data_inputs(
        self, data_set: str | pd.DataFrame
    ) -> tuple[torch.Tensor, pd.DataFrame, pd.DataFrame, bool]:
        """Process a new data set of inputs and format them for a surrogate model to use

        The new data may be incomplete. The function expects a long data table (unpivotted). This function is under-optimized, and should not be used during training.

        Args:
            data_set (str | pd.DataFrame):
            Either "training" or "validation" OR
            An input data frame on which to predict with the GP. Should contain the following columns
            - `id`
            - one column per descriptor
            - `protocol_name`
            - `value` (Optional)

        Returns:
            torch.Tensor: the inputs to provide to a surrogate model for predicting the same values as provided in the data set
            pd.DataFrame: the processed data frame, in a wide format
            pd.DataFrame: the original data frame, in a long format
            bool: a flag, True if the value column is dummy in the output data frames
        """
        if isinstance(data_set, str):
            if data_set == "training":
                patients = self.training_patients
            elif data_set == "validation":
                patients = self.validation_patients
            else:
                raise ValueError(
                    f"Incorrect data set choice: {data_set}. Use `training` or `validation`"
                )
            new_data = self.full_df_raw.loc[self.full_df_raw["id"].isin(patients)]
        elif isinstance(data_set, pd.DataFrame):
            new_data = data_set
        else:
            raise ValueError(
                "`predict_new_data` expects either a str (`training`|`validation`) or a data frame."
            )

        # Validate the content of the new data frame
        new_columns = new_data.columns.to_list()
        if not "protocol_arm" in new_columns:
            new_protocols = ["identity"]
        else:
            new_protocols = new_data["protocol_arm"].unique().tolist()
        new_outputs = new_data["output_name"].unique().tolist()
        if not (set(new_protocols) <= set(self.protocol_arms)):
            raise ValueError(
                "Supplied data frame contains a different set of protocol arms."
            )
        if not (set(new_outputs) <= set(self.output_names)):
            raise ValueError(
                "Supplied data frame contains a different set of model outputs."
            )
        if not (set(self.parameter_names) <= set(new_columns)):
            raise ValueError(
                "All model descriptors are not supplied in the new data frame."
            )

        # Flag the case where no observed value was supplied
        remove_value = False
        if not "value" in new_columns:
            remove_value = True
            # Add a dummy `value` column
            new_data["value"] = 1.0

        wide_df = self.pivot_input_data(new_data)
        tensor_inputs_wide = torch.as_tensor(
            wide_df[self.parameter_names].values, device=device
        )

        return tensor_inputs_wide, wide_df, new_data, remove_value

    def merge_predictions_long(
        self,
        pred: tuple[torch.Tensor, torch.Tensor],
        wide_df: pd.DataFrame,
        long_df: pd.DataFrame,
        remove_value: bool,
    ) -> pd.DataFrame:
        """Merge model predictions with an observation data frame

        Args:
            pred (tuple[torch.Tensor, torch.Tensor, torch.Tensor]): Predictions from a model: mean, low bound, high bound. Predictions are expected in a wide format (as many columns as tasks)
            wide_df (pd.DataFrame): The comparison data frame, in a wide format
            long_df (pd.DataFrame): The comparison data frame, in a long format
            remove_value (bool): True if the value column should be ignored

        Returns:
            pd.DataFrame: A merged data frame in a long format, identical to the initial data, with additional columns [`pred_mean`, `pred_var`, `pred_low`, `pred_high`]
        """
        pred_mean, pred_variance = pred
        # Reshape these outputs into a long format
        mean_df = self.pivot_outputs_longer(wide_df, pred_mean, "pred_mean")
        var_df = self.pivot_outputs_longer(wide_df, pred_variance, "pred_var")
        # Merge the model results with the long format data frame
        out_df = reduce(
            lambda left, right: pd.merge(
                left,
                right,
                on=["id"] + self.parameter_names + ["protocol_arm", "output_name"],
                how="left",
            ),
            [long_df, mean_df, var_df],
        )
        out_df["pred_low"] = out_df.apply(
            lambda r: r["pred_mean"] - 2 * np.sqrt(r["pred_var"]), axis=1
        )
        out_df["pred_high"] = out_df.apply(
            lambda r: r["pred_mean"] + 2 * np.sqrt(r["pred_var"]), axis=1
        )
        # Remove the dummy value column if it was added during the data processing
        if remove_value:
            out_df = out_df.drop(columns=["value"])
        return out_df



import pytest
import numpy as np
import matplotlib.pyplot as plt


@pytest.fixture(scope="session")
def np_rng():
    # Initialize the seeds for all random operators used in the tests
    rng = np.random.default_rng(42)
    return rng


@pytest.fixture(autouse=True)
def clean_matplotlib_figures():
    """
    Automatically closes all matplotlib figures after each test.
    """
    yield  # Run the test

    # Teardown: Close all open figures
    plt.close("all")



from matplotlib import pyplot as plt
import math
import torch
import gpytorch
from tqdm.notebook import tqdm
from gpytorch.mlls import VariationalELBO, PredictiveLogLikelihood
from torch.utils.data import TensorDataset, DataLoader
import numpy as np
import pandas as pd
from typing import Optional, cast

from .data import TrainingDataSet
from .plot import (
    plot_all_solutions,
    plot_individual_solution,
    plot_obs_vs_predicted,
    LossPlot,
)
from ..utils import smoke_test, device

torch.set_default_dtype(torch.float64)
gpytorch.settings.cholesky_jitter(1e-6)


class SVGP(gpytorch.models.ApproximateGP):
    """The internal GP class used to create surrogate models, interfacing with gpytorch's API"""

    def __init__(
        self,
        inducing_points: torch.Tensor,
        nb_params: int,
        nb_tasks: int,
        nb_latents: int,
        var_dist: str = "Chol",
        var_strat: str = "IMV",
        kernel: str = "RBF",
        deep_kernel: bool = False,
        jitter: float = 1e-6,
        nb_mixtures: int = 4,  # only for the SMK kernel
        nb_features: int = 10,
    ):
        """_summary_

        Args:
            inducing_points (torch.Tensor): Initial choice for the inducing points
            nb_params (int): Number of input parameters
            nb_outputs (int): Number of outputs (tasks)
            var_dist (str, optional): Variational distribution choice. Defaults to "Chol".
            var_strat (str, optional): Variational strategy choice. Defaults to "IMV".
            kernel (str, optional): Kernel choice. Defaults to "RBF".
            deep_kernel (bool, optional): Add a neural network feature extractor in the kernel
            jitter (float, optional): Jitter value (for numerical stability). Defaults to 1e-6.
            nb_mixtures (int, optional): Number of mixtures for the SMK kernel. Defaults to 4.
            nb_features (int, optional): Number of features for the deep kernel. Defaults to 10.
        """
        assert var_dist == "Chol", f"Unsupported variational distribution: {var_dist}"
        if var_strat == "LMCV":
            self.batch_size = nb_latents
        elif var_strat == "IMV":
            self.batch_size = nb_tasks
        else:
            self.batch_size = nb_tasks

        self.kernel_type = kernel
        self.deep_kernel = deep_kernel
        if deep_kernel:
            self.kernel_size = nb_features
        else:
            self.kernel_size = nb_params

        variational_distribution = gpytorch.variational.CholeskyVariationalDistribution(
            inducing_points.shape[0],
            batch_shape=torch.Size([self.batch_size]),
            mean_init_std=1e-3,
        )
        if var_strat == "IMV":
            variational_strategy = (
                gpytorch.variational.IndependentMultitaskVariationalStrategy(
                    gpytorch.variational.VariationalStrategy(
                        self,
                        inducing_points,
                        variational_distribution,
                        learn_inducing_locations=True,
                        jitter_val=jitter,
                    ),
                    num_tasks=nb_tasks,
                )
            )
        elif var_strat == "LMCV":
            variational_strategy = gpytorch.variational.LMCVariationalStrategy(
                gpytorch.variational.VariationalStrategy(
                    self,
                    inducing_points,
                    variational_distribution,
                    learn_inducing_locations=True,
                    jitter_val=jitter,
                ),
                num_tasks=nb_tasks,
                num_latents=nb_latents,
                latent_dim=-1,
            )
        else:
            raise ValueError(f"Unsupported variational strategy {var_strat}")

        super().__init__(variational_strategy)

        # Todo : allow for different mean choices
        self.mean_module = gpytorch.means.ConstantMean(
            batch_shape=torch.Size([self.batch_size])
        )

        if kernel == "RBF":
            self.covar_module = gpytorch.kernels.ScaleKernel(
                gpytorch.kernels.RBFKernel(
                    batch_shape=torch.Size([self.batch_size]),
                    ard_num_dims=self.kernel_size,
                    jitter=jitter,
                ),
                batch_shape=torch.Size([self.batch_size]),
            )
        elif kernel == "SMK":
            self.covar_module = gpytorch.kernels.SpectralMixtureKernel(
                batch_size=self.batch_size,
                num_mixtures=nb_mixtures,
                ard_num_dims=self.kernel_size,
                jitter=jitter,
            )
        elif kernel == "Matern":
            self.covar_module = gpytorch.kernels.ScaleKernel(
                gpytorch.kernels.MaternKernel(
                    nu=2.5,
                    batch_size=nb_tasks,
                    num_mixtures=nb_mixtures,
                    ard_num_dims=self.kernel_size,
                    jitter=jitter,
                ),
                batch_shape=torch.Size([self.batch_size]),
            )
        else:
            raise ValueError(f"Unsupported kernel {kernel}")
        if deep_kernel:
            self.feature_extractor = LargeFeatureExtractor(nb_params, nb_features)
            self.scale_to_bounds = gpytorch.utils.grid.ScaleToBounds(-1.0, 1.0)

    def forward(self, x: torch.Tensor):
        if self.deep_kernel:
            proj_x = self.feature_extractor(x)
            proj_x = self.scale_to_bounds(proj_x)
            mean_x = cast(torch.Tensor, self.mean_module(proj_x))
            covar_x = self.covar_module(proj_x)
        else:
            mean_x = cast(torch.Tensor, self.mean_module(x))
            covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


class GP:
    """GP surrogate model"""

    def __init__(
        self,
        training_df: pd.DataFrame,
        descriptors: list[str],
        var_dist: str = "Chol",  # only Cholesky currently supported
        var_strat: str = "IMV",  # either IMV (Independent Multitask Variational) or LMCV (Linear Model of Coregionalization Variational)
        kernel: str = "RBF",  # RBF or SMK
        deep_kernel: bool = True,
        nb_training_iter: int = 400,
        min_delta: float = 0.05,
        patience: int = 50,
        training_proportion: float = 0.7,
        nb_inducing_points: int = 200,
        data_already_normalized: bool = False,
        log_lower_limit: float = 1e-10,
        log_inputs: list[str] = [],
        log_outputs: list[str] = [],
        nb_latents: Optional[int] = None,
        # by default we will use nb_latents = nb_outputs
        mll: str = "ELBO",  # ELBO or PLL
        learning_rate: Optional[float] = None,  # optional
        lr_decay: Optional[float] = None,
        num_mixtures: int = 4,  # only for the SMK kernel
        nb_features: int = 10,  # only for the DK kernel
        jitter: float = 1e-6,
        plot_frame_rate: int = 10,
    ):
        """Instantiate a GP model on a training data frame

        Args:
            training_df (pd.DataFrame): Training data frame containing the following columns:
              - `id`: the id of the patient, str or int
              - `descriptors`: one column per patient descriptor (including `time`, if necessary)
              - `output_name`: the name of simulated model output
              - `value`: the simulated value (for a given patient, protocol arm and output name)
              - `protocol_arm` (optional): the protocol arm on which this patient was simulated. If not provided, `identity` will be used
            descriptors (list[str]): the names of the columns of `training_df` which correspond to descriptors on which to train the GP
            var_dist (str, optional): Variational distribution choice. Defaults to "Chol".
            nb_training_iter (int, optional): Number of iterations for training. Defaults to 400.
            training_proportion (float, optional): Proportion of patients to be used as training vs. validation. Defaults to 0.7.
            nb_inducing_points (int, optional): Number of inducing points to be used for variational inference. Defaults to 200.
            log_inputs (list[str]): the list of parameter inputs which should be rescaled to log when fed to the GP. Avoid adding time here, or any parameter that takes 0 as a value.
            log_outputs (list[str]): list of model outptus which should be rescaled to log
            log_lower_limit(float): epsilon value that is added to all rescaled value to avoid numerical errors when log-scaling variables
            nb_latents (Optional[int], optional): Number of latents. Defaults to None, implying that nb_latents = nb_tasks will be used
            mll (str, optional): Marginal log likelihood choice. Defaults to "ELBO" (other option "PLL")
            learning_rate (Optional[float]): learning rate initial value. Defaults to 0.001 (in torch.optim.Adam)
            lr_decay (Optional[float]): learning rate decay rate.
            num_mixtures (int): Number of mixtures used in the SMK kernel. Not used for other kernel choices. Default to 4.
            nb_features (int): Number of features used in the deep kernel. Not used for other kernel choices. Default to 10.
            jitter: Jitter value for numerical stability

        Comments:
            The GP will learn nb_tasks = nb_outputs * nb_protocol_arms, i.e. one predicted task per model output per protocol arm.

        """
        # Define GP parameters
        self.var_dist = var_dist
        self.var_strat = var_strat
        self.kernel = kernel
        self.deep_kernel = deep_kernel
        if smoke_test:
            self.nb_training_iter = 1
            self.nb_inducing_points = 10
        else:
            self.nb_training_iter = nb_training_iter
            self.nb_inducing_points = nb_inducing_points
        self.learning_rate = learning_rate
        self.mll_name = mll
        self.num_mixtures = num_mixtures
        self.nb_features = nb_features
        self.jitter = jitter
        if lr_decay is not None:
            self.lr_decay = lr_decay

        self.min_delta = min_delta
        self.patience = patience
        self.min_loss = np.inf
        self.plot_frame_rate = plot_frame_rate

        self.data = TrainingDataSet(
            training_df,
            descriptors,
            training_proportion,
            log_lower_limit,
            log_inputs,
            log_outputs,
            data_already_normalized,
        )

        if nb_latents is None:
            self.nb_latents = self.data.nb_tasks
        else:
            self.nb_latents = nb_latents
        # Create inducing points
        self.inducing_points = self.data.X_training[
            torch.randperm(self.data.X_training.shape[0])[: self.nb_inducing_points],
            :,
        ]

        # Initialize likelihood and model
        self.likelihood = gpytorch.likelihoods.MultitaskGaussianLikelihood(
            num_tasks=self.data.nb_tasks, has_global_noise=True, has_task_noise=True
        )
        self.model = SVGP(
            inducing_points=self.inducing_points,
            nb_params=self.data.nb_parameters,
            nb_tasks=self.data.nb_tasks,
            nb_latents=self.nb_latents,
            var_dist=self.var_dist,
            var_strat=self.var_strat,
            kernel=self.kernel,
            jitter=self.jitter,
            nb_mixtures=self.num_mixtures,
            nb_features=self.nb_features,
        )

        # set the marginal log likelihood
        if self.mll_name == "ELBO":
            self.mll = VariationalELBO(
                self.likelihood, self.model, num_data=self.data.Y_training.size(0)
            )
        elif self.mll_name == "PLL":
            self.mll = PredictiveLogLikelihood(
                self.likelihood, self.model, num_data=self.data.Y_training.size(0)
            )
        else:
            raise ValueError(f"Invalid MLL choice ({self.mll}). Choose ELBO or PLL.")

        # Move all components to the selected device
        self.model.to(device)
        self.likelihood.to(device)
        self.mll.to(device)

    def train(
        self, mini_batching: bool = False, mini_batch_size: Optional[int] = None
    ) -> None:
        # set model and likelihood in training mode
        self.model.train()
        self.likelihood.train()

        # initialize the adam optimizer
        params_to_optim = [
            {"params": self.model.parameters()},
            {"params": self.likelihood.parameters()},
        ]
        if self.learning_rate is None:
            optimizer = torch.optim.Adam(params_to_optim)
        else:
            optimizer = torch.optim.Adam(params_to_optim, lr=self.learning_rate)
        if hasattr(self, "lr_decay"):
            scheduler = torch.optim.lr_scheduler.ExponentialLR(
                optimizer, gamma=self.lr_decay
            )
        else:
            scheduler = None

        # keep track of the loss
        self.training_losses = []
        self.validation_losses = []

        self.loss_plot = LossPlot()

        epochs = tqdm(range(self.nb_training_iter), desc="Epochs", position=0)
        with gpytorch.settings.observation_nan_policy("fill"):
            if mini_batching:
                # set the mini_batch_size to a power of two of the total size -4
                if mini_batch_size == None:
                    power = np.maximum(
                        math.floor(math.log2(self.data.X_training.shape[0])) - 4, 1
                    )
                    self.mini_batch_size: int | None = math.floor(2**power)
                else:
                    self.mini_batch_size = mini_batch_size
            else:
                # No batching - load the whole data set at each iteration
                self.mini_batch_size = self.data.X_training.shape[0]

            # prepare data loader
            train_dataset = TensorDataset(self.data.X_training, self.data.Y_training)
            train_loader = DataLoader(
                train_dataset,
                batch_size=self.mini_batch_size,
                shuffle=True,
            )

            # main training loop
            for i in epochs:
                epoch_loss = 0.0
                for batch_params, batch_outputs in tqdm(
                    train_loader,
                    desc="Batch progress",
                    position=1,
                    leave=False,
                    disable=not mini_batching,
                ):
                    optimizer.zero_grad()  # zero gradients from previous iteration
                    output = self.model(batch_params)  # recalculate the prediction
                    loss = -cast(torch.Tensor, self.mll(output, batch_outputs))
                    loss.backward()  # compute the gradients of the parameters that can be changed
                    optimizer.step()
                    epoch_loss += loss.item()
                epoch_loss = epoch_loss / len(train_loader)
                self.training_losses.append(epoch_loss)
                if scheduler is not None:
                    scheduler.step()
                # In case we have a validation data set
                if self.data.training_proportion < 1.0:
                    validation_loss = -cast(
                        torch.Tensor,
                        self.mll(
                            self.model(self.data.X_validation),
                            self.data.Y_validation,
                        ),
                    ).item()
                    # Append validation loss
                    self.validation_losses.append(validation_loss)
                    epochs.set_postfix(
                        {
                            "training loss": epoch_loss,
                            "validation loss": validation_loss,
                        }
                    )
                    # Define the loss that will be tested for convergence
                    current_loss = float(validation_loss)
                    if i > 0:
                        previous_loss = self.validation_losses[-2]
                    else:
                        previous_loss = np.inf
                else:
                    current_loss = float(epoch_loss)
                    if i > 0:
                        previous_loss = self.training_losses[-2]
                    else:
                        previous_loss = np.inf
                    epochs.set_postfix(
                        {
                            "training loss": epoch_loss,
                        }
                    )
                if self.early_stop(current_loss, previous_loss):
                    # Break out of the iterations
                    print(f"Early convergence reached, stopping after {i} iterations.")
                    # Update loss plot one final time
                    self.loss_plot.update_plot(
                        np.arange(i + 1),
                        np.array(self.training_losses),
                        np.array(self.validation_losses),
                    )
                    plt.close(self.loss_plot.fig)
                    break
                # Update loss plot at frame rate
                if (i % self.plot_frame_rate == 0) | (i == self.nb_training_iter - 1):
                    self.loss_plot.update_plot(
                        np.arange(i + 1),
                        np.array(self.training_losses),
                        np.array(self.validation_losses),
                    )
            plt.close(self.loss_plot.fig)

    def early_stop(self, current_loss: float, previous_loss: float) -> bool:
        """Early convergence criterion

        Returns a stop signal when the loss is stabilized for more than a certain number of iterations (self.patience)

        Args:
            current_loss (float): The loss from current iteration
            previous_loss (float): The loss from previous iteration

        Returns:
            bool: True is convergence criterion is met
        """
        converged = False
        if np.abs(current_loss - previous_loss) > self.min_delta:
            self.patience_counter = 0
        else:
            self.patience_counter += 1
            if self.patience_counter >= self.patience:
                converged = True
        return converged

    def _predict_training(
        self, X: torch.Tensor
    ) -> tuple[torch.Tensor, tuple[torch.Tensor, torch.Tensor]]:
        """Internal method used to predict normalized outputs on normalized inputs."""
        self.model.eval()
        self.likelihood.eval()

        with torch.no_grad():
            pred = cast(
                gpytorch.distributions.MultitaskMultivariateNormal,
                self.likelihood(self.model(X)),
            )

        return pred.mean, pred.confidence_region()

    def predict_wide_scaled(self, X: torch.Tensor) -> tuple[torch.Tensor, torch.Tensor]:
        """Predict mean and interval confidence values for a given input tensor (normalized inputs). This function outputs rescaled values."""
        self.model.eval()
        self.likelihood.eval()
        inputs = self.data.normalize_inputs_tensor(X)

        with torch.no_grad():
            pred = self.model(inputs)

        if self.data.data_already_normalized:
            out_mean = pred.mean
        else:
            out_mean = self.data.unnormalize_output_wide(pred.mean)
        return out_mean, pred.variance

    def predict_long_scaled(
        self, X: torch.Tensor, tasks: torch.LongTensor
    ) -> tuple[torch.Tensor, torch.Tensor]:
        """Predict outputs from the GP in a long format (one row per task)"""

        self.model.eval()
        self.likelihood.eval()
        inputs = self.data.normalize_inputs_tensor(X)
        with torch.no_grad():
            pred = self.model(inputs, task_indices=tasks)
        if self.data.data_already_normalized:
            out_mean = pred.mean
        else:
            out_mean = self.data.unnormalize_output_long(pred.mean, task_indices=tasks)
        return out_mean, pred.variance

    def RMSE(self, y1: torch.Tensor, y2: torch.Tensor) -> torch.Tensor:
        """Given two tensors of same shape, compute the Root Mean Squared Error on each column (outputs)."""
        assert y1.shape == y2.shape
        # Ignore potential NaN values in the RMSE computation
        mask = (~torch.isnan(y1)) * (~torch.isnan(y2))
        squared_residuals = torch.where(mask, torch.pow(y1 - y2, 2), 0)
        return torch.sqrt(squared_residuals.sum(dim=0) / mask.sum(dim=0))

    def eval_perf(self):
        """Evaluate the model performance on its training data set and validation data set (normalized inputs and ouptuts)."""

        def print_task_rmse(index, val):
            print(
                f"    Output: {self.data.output_names[self.data.task_idx_to_output_idx[index]]}, protocol: {self.data.task_idx_to_protocol[index]}, RMSE: {val:.4f}"
            )

        (
            self.Y_training_predicted_mean,
            _,
        ) = self._predict_training(self.data.X_training)
        self.RMSE_training = self.RMSE(
            self.Y_training_predicted_mean, self.data.Y_training
        )
        print("Training data set:")

        for i, err in enumerate(self.RMSE_training):
            print_task_rmse(i, err.item())

        if not (self.data.X_validation is None) and not (
            self.data.Y_validation is None
        ):
            (
                self.Y_validation_predicted_mean,
                _,
            ) = self._predict_training(self.data.X_validation)
            self.RMSE_validation = self.RMSE(
                self.Y_validation_predicted_mean, self.data.Y_validation
            )
            print("Validation data set:")
            for i, err in enumerate(self.RMSE_validation):
                print_task_rmse(i, err.item())

    def predict_new_data(self, data_set: str | pd.DataFrame) -> pd.DataFrame:
        """Process a new data set of inputs and predict using the GP

        The new data may be incomplete. The function expects a long data table (unpivotted). This function is under-optimized, and should not be used during training.

        Args:
            data_set (str | pd.DataFrame):
            Either "training" or "validation" OR
            An input data frame on which to predict with the GP. Should contain the following columns
            - `id`
            - one column per descriptor
            - `protocol_name`

        Returns:
            pd.DataFrame: Same data frame as new_data, with additional columns
            - `pred_mean`
            - `pred_low`
            - `pred_high`
        """
        # Fetch the data
        X_wide, wide_df, long_df, remove_value = self.data.get_data_inputs(data_set)

        # Simulate the model - using a wide output (all tasks predicted for each observation)
        pred = self.predict_wide_scaled(X_wide)
        out_df = self.data.merge_predictions_long(pred, wide_df, long_df, remove_value)
        return out_df

    def plot_obs_vs_predicted(
        self,
        data_set: pd.DataFrame | str,
        fig_scaling: tuple[float, float] = (5.0, 2.0),
        logScale=None,
    ):
        """Plots the observed vs. predicted values on the training or validation data set, or on a new data set."""

        obs_vs_pred = self.predict_new_data(data_set)
        plot_obs_vs_predicted(obs_vs_pred, fig_scaling, logScale)

    # plot function
    def plot_individual_solution(
        self, patient_number: int, fig_scaling: tuple[float, float] = (5.0, 3.0)
    ):
        """Plot the model prediction (and confidence interval) vs. the input data for a single patient from the GP's internal data set. Can be either training or validation patient."""
        patient_index = self.data.patients[patient_number]
        input_df = self.data.full_df_raw.loc[
            self.data.full_df_raw["id"] == patient_index
        ]
        obs_vs_pred = self.predict_new_data(input_df)
        plot_individual_solution(obs_vs_pred, fig_scaling)

    def plot_all_solutions(
        self,
        data_set: str | pd.DataFrame,
        fig_scaling: tuple[float, float] = (5.0, 2.0),
    ):
        """Plot the overlapped observations and model predictions for all patients, either on one the GP's intrinsic data sets, or on a new data set."""

        obs_vs_pred = self.predict_new_data(data_set)
        plot_all_solutions(obs_vs_pred, fig_scaling)


class LargeFeatureExtractor(torch.nn.Sequential):
    def __init__(self, n_params, n_features):
        super(LargeFeatureExtractor, self).__init__()
        self.add_module("linear1", torch.nn.Linear(n_params, 1000))
        self.add_module("relu1", torch.nn.ReLU())
        self.add_module("linear2", torch.nn.Linear(1000, 500))
        self.add_module("relu2", torch.nn.ReLU())
        self.add_module("linear3", torch.nn.Linear(500, 50))
        self.add_module("relu3", torch.nn.ReLU())
        self.add_module("linear4", torch.nn.Linear(50, n_features))



import pytest
import numpy as np
import pandas as pd

from vpop_calibration import *


@pytest.fixture
def ode_model_setup():
    def equations_with_abs(t, y, k_a, k_12, k_21, k_el):
        A_absorption, A_central, A_peripheral = y[0], y[1], y[2]
        dA_absorption_dt = -k_a * A_absorption
        dA_central_dt = (
            k_a * A_absorption
            + k_21 * A_peripheral
            - k_12 * A_central
            - k_el * A_central
        )
        dA_peripheral_dt = k_12 * A_central - k_21 * A_peripheral

        ydot = [dA_absorption_dt, dA_central_dt, dA_peripheral_dt]
        return ydot

    variable_names = ["A0", "A1", "A2"]
    parameter_names = ["k_a", "k_12", "k_21", "k_el"]

    protocol_design = pd.DataFrame(
        {"protocol_arm": ["arm-A", "arm-B"], "k_el": [0.5, 10.0]}
    )

    initial_conditions = np.array([10.0, 0.0, 0.0])

    return (
        equations_with_abs,
        variable_names,
        parameter_names,
        protocol_design,
        initial_conditions,
    )


@pytest.fixture
def param_structure(use_case):
    match use_case:
        case 1:
            log_mi = {}
            log_pdu = {
                "k_12": {"mean": -1.0, "sd": 0.25},
                "k_21": {"mean": -1.0, "sd": 0.25},
                "k_a": {"mean": -1.0, "sd": 0.25},
            }
        case _:
            log_mi = {"k_21": 0.0}
            log_pdu = {
                "k_12": {"mean": -1.0, "sd": 0.25},
                "k_a": {"mean": -1.0, "sd": 0.25},
            }
    return log_mi, log_pdu


@pytest.fixture
def covariate_map_for_tests(include_cov):
    if include_cov:
        cov_map = {
            "k_12": {"foo": {"coef": "cov_foo_k12", "value": 0.2}},
            "k_21": {},
            "k_a": {},
        }
    else:
        cov_map = None
    return cov_map


@pytest.fixture
def patients_df_for_tests(include_cov):
    patients_df = pd.DataFrame({"id": ["p1", "p2"], "protocol_arm": ["arm-A", "arm-B"]})
    if include_cov:
        patients_df["foo"] = [1.0, 2.0]
    return patients_df


@pytest.mark.parametrize("error_model", ["additive", "proportional"])
@pytest.mark.parametrize("include_cov", [True, False])
@pytest.mark.parametrize("use_case", [1, 2])
def test_generate_data_omega(
    ode_model_setup,
    patients_df_for_tests,
    covariate_map_for_tests,
    param_structure,
    error_model,
):
    (
        equations,
        variable_names,
        parameter_names,
        protocol_design,
        initial_conditions,
    ) = ode_model_setup
    log_mi, log_pdu = param_structure

    pk_model = OdeModel(equations, variable_names, parameter_names, multithreaded=False)
    time_steps = [0.0, 1.0]
    # Parameter definitions
    true_res_var = [0.5, 0.02, 0.1]
    time_steps = np.arange(0.0, 10.0, 4.0)
    patients_df = patients_df_for_tests
    covariate_map = covariate_map_for_tests

    obs_df = simulate_dataset_from_omega(
        pk_model,
        protocol_design,
        time_steps,
        initial_conditions,
        log_mi,
        log_pdu,
        error_model,
        true_res_var,
        covariate_map,
        patients_df,
    )



import pytest
import torch
import pandas as pd
from torch.testing import assert_close

from vpop_calibration import *


def test_analytical_saem(np_rng):
    def logistic_growth(lambda1, lambda2, lambda3, t):
        y = lambda1 / (1 + torch.exp(-(t - lambda2) / lambda3))
        return y

    outputs = ["circ"]
    struct_model = StructuralAnalytical(logistic_growth, outputs)

    patient_df = pd.DataFrame({"id": ["tree-1", "tree-2"]})
    df = patient_df.merge(pd.DataFrame({"output_name": outputs}), how="cross")
    time_steps = [0.0, 1.0, 3.0]
    df = df.merge(pd.DataFrame({"time": time_steps}), how="cross")
    df["value"] = np_rng.normal(0, 1)

    init_log_mi = {"lambda2": 0.0, "lambda3": 0.0}
    init_log_pdu = {
        "lambda1": {"mean": 0.0, "sd": 0.5},
    }
    constraints = {"lambda1": {"low": 0.0, "high": 300}}
    covariate_map = None
    init_res_var = [100.0]
    nlme_model = NlmeModel(
        structural_model=struct_model,
        patients_df=patient_df,
        init_log_MI=init_log_mi,
        init_PDU=init_log_pdu,
        covariate_map=covariate_map,
        init_res_var=init_res_var,
        error_model_type="additive",
        num_chains=1,
        constraints=constraints,
    )

    optimizer = PySaem(nlme_model, df)
    optimizer.run()
    nlme_model.compute_ebe()
    nlme_model.sample_conditional_distribution(nb_samples=2)
    plot_individual_map_estimates(nlme_model)
    plot_all_individual_map_estimates(nlme_model)
    plot_map_estimates_gof(nlme_model)
    plot_weighted_residuals(nlme_model, "iwres")
    plot_weighted_residuals(nlme_model, "pwres")
    plot_weighted_residuals(nlme_model, "npde")


def test_analytical_no_protocol():
    def logistic_growth(lambda1, lambda2, lambda3, t):
        return lambda1 / (1 + torch.exp(-(t - lambda2) / lambda3))

    struct_model = StructuralAnalytical(
        logistic_growth,
        ["circumference"],
        protocol_design=None,
    )

    assert struct_model.parameter_names == ["lambda1", "lambda2", "lambda3"]
    assert struct_model.nb_protocol_overrides == 0
    assert struct_model.tasks == ["circumference_identity"]
    assert struct_model.task_idx_to_protocol == {0: "identity"}
    assert struct_model.task_idx_to_output_idx == {0: 0}
    num_tasks = len(struct_model.tasks)
    num_protocol_parameters = 0
    assert struct_model.task_protocol_tensor.shape == (
        num_tasks,
        num_protocol_parameters,
    )


def test_analytical_one_arm_one_override():
    def logistic_growth(lambda1, lambda2, lambda3, t, sun_power_multiplicator):
        return lambda1 / (
            1 + torch.exp(-(sun_power_multiplicator * (t - lambda2) / lambda3))
        )

    struct_model = StructuralAnalytical(
        logistic_growth,
        ["circumference"],
        protocol_design=pd.DataFrame(
            data={"protocol_arm": ["Italy"], "sun_power_multiplicator": [1.05]}
        ),
    )

    assert struct_model.parameter_names == [
        "lambda1",
        "lambda2",
        "lambda3",
    ]
    assert struct_model.nb_protocol_overrides == 1
    assert struct_model.tasks == ["circumference_Italy"]
    assert struct_model.task_idx_to_protocol == {0: "Italy"}
    assert struct_model.task_idx_to_output_idx == {0: 0}
    num_tasks = len(struct_model.tasks)
    num_protocol_parameters = 1
    assert struct_model.task_protocol_tensor.shape == (
        num_tasks,
        num_protocol_parameters,
    )


def test_analytical_one_arm_two_overrides():
    # Here we scrambled the function arguments on purpose, to check
    # that the logic of self.input_tensor_column_index_to_function_parameter_index
    # is properly implemented
    def logistic_growth(
        lambda3,
        t,
        max_circumference_multiplicator,
        lambda2,
        lambda1,
        sun_power_multiplicator,
    ):
        return (
            max_circumference_multiplicator
            * lambda1
            / (1 + torch.exp(-(sun_power_multiplicator * (t - lambda2) / lambda3)))
        )

    struct_model = StructuralAnalytical(
        logistic_growth,
        ["circumference"],
        protocol_design=pd.DataFrame(
            data={
                "protocol_arm": ["Italy"],
                "sun_power_multiplicator": [1.05],
                "max_circumference_multiplicator": [1.1],
            }
        ),
    )

    assert struct_model.parameter_names == [
        "lambda3",
        "lambda2",
        "lambda1",
    ]

    # internally we have input_parameters = parameter_names_without_protocol_overrides + ["t"] + protocol_parameters
    # ["lambda3", "lambda2", "lambda1", "t", "max_circumference_multiplicator", "sun_power_multiplicator"]
    #      0         1           2       3                    4                          5
    assert struct_model.input_tensor_column_index_to_function_parameter_index == [
        0,  # position of "lambda3"
        3,  # position of "t"
        4,  # position of "max_circumference_multiplicator"
        1,  # position of "lambda2"
        2,  # position of "lambda1"
        5,  # position of "sun_power_multiplicator"
    ]
    assert struct_model.nb_protocol_overrides == 2
    assert struct_model.tasks == ["circumference_Italy"]
    assert struct_model.task_idx_to_protocol == {0: "Italy"}
    assert struct_model.task_idx_to_output_idx == {0: 0}
    num_tasks = len(struct_model.tasks)
    num_protocol_parameters = 2
    assert struct_model.task_protocol_tensor.shape == (
        num_tasks,
        num_protocol_parameters,
    )


def test_analytical_two_arms_two_overrides():
    def logistic_growth(
        lambda1,
        lambda2,
        lambda3,
        t,
        sun_power_multiplicator,
        max_circumference_multiplicator,
    ):
        return (
            max_circumference_multiplicator
            * lambda1
            / (1 + torch.exp(-(sun_power_multiplicator * (t - lambda2) / lambda3)))
        )

    protocol_design = pd.DataFrame(
        data={
            "protocol_arm": ["Italy", "Morocco"],
            "sun_power_multiplicator": [1.05, 1.2],
            "max_circumference_multiplicator": [1.1, 1.25],
        }
    )
    struct_model = StructuralAnalytical(
        logistic_growth,
        ["circumference"],
        protocol_design=protocol_design,
    )

    assert struct_model.parameter_names == [
        "lambda1",
        "lambda2",
        "lambda3",
    ]
    assert struct_model.nb_protocol_overrides == 2
    assert struct_model.tasks == ["circumference_Italy", "circumference_Morocco"]
    assert struct_model.task_idx_to_protocol == {0: "Italy", 1: "Morocco"}
    assert struct_model.task_idx_to_output_idx == {0: 0, 1: 0}
    num_tasks = len(struct_model.tasks)
    num_protocol_parameters = 2
    assert struct_model.task_protocol_tensor.shape == (
        num_tasks,
        num_protocol_parameters,
    )
    patient_index = torch.tensor([0, 0, 0, 1, 1, 1])
    timestep_index = torch.tensor([0, 1, 2, 0, 1, 2])
    #                           Italy    Morocco
    task_index = torch.tensor([0, 0, 0, 1, 1, 1])
    X = torch.tensor(
        [
            [
                [  # lam1 lam2 lam3  t
                    [333, 101, 350, 118],  # patient 1 at t1
                    [333, 101, 350, 484],  # patient 1 at t2
                    [333, 101, 350, 664],  # patient 1 at t3
                ],
                [
                    [483, 100, 360, 118],  # patient 2 at t1
                    [483, 100, 360, 484],  # patient 2 at t2
                    [483, 100, 360, 664],  # patient 2 at t3
                ],
            ]
        ]
    )
    (num_chains, nb_patients, nb_timesteps, nb_params) = X.shape
    actual_y, _pred_var = struct_model.simulate(
        X, (patient_index, timestep_index, task_index), chunks=None
    )
    expected_y = logistic_growth(
        lambda1=torch.tensor([333, 333, 333, 483, 483, 483]),
        lambda2=torch.tensor([101, 101, 101, 100, 100, 100]),
        lambda3=torch.tensor([350, 350, 350, 360, 360, 360]),
        t=torch.tensor([118, 484, 664, 118, 484, 664]),
        #                                           Italy          Morocco
        sun_power_multiplicator=torch.tensor([1.05, 1.05, 1.05, 1.2, 1.2, 1.2]),
        #                                                  Italy          Morocco
        max_circumference_multiplicator=torch.tensor([1.1, 1.1, 1.1, 1.25, 1.25, 1.25]),
    ).reshape((num_chains, nb_patients * nb_timesteps))
    assert_close(actual_y, expected_y)



import pandas as pd

from vpop_calibration import *


def test_gp_saem(np_rng):
    time_steps = pd.DataFrame({"time": [0.0, 1.0]})
    params = ["k1", "k2", "k3", "time"]
    patients_training = pd.DataFrame(
        {"id": ["p1", "p2"], "k1": [1.0, 2.0], "k2": [3.0, 4.0], "k3": [5.0, 6.0]}
    )
    outputs = pd.DataFrame({"output_name": ["s1", "s2"]})
    protocol_arms = pd.DataFrame({"protocol_arm": ["arm-A", "arm-B"]})

    training_df = (
        patients_training.merge(outputs, how="cross")
        .merge(protocol_arms, how="cross")
        .merge(time_steps, how="cross")
    )
    training_df["value"] = np_rng.normal()

    gp = GP(training_df, params)

    patient_df = pd.DataFrame(
        {"id": ["patient-1", "patient-2"], "protocol_arm": ["arm-B", "arm-A"]}
    )
    obs_df = patient_df.merge(outputs, how="cross").merge(time_steps, how="cross")
    obs_df["value"] = np_rng.normal()
    obs_df = obs_df.sample(frac=0.7)

    init_mi = {"k1": 0.0}
    init_pdu = {"k2": {"mean": 0.0, "sd": 0.1}, "k3": {"mean": 0.0, "sd": 0.1}}
    init_res_var = [0.5, 0.5]

    struct_model = StructuralGp(gp)

    nlme_model = NlmeModel(
        struct_model,
        patient_df,
        init_mi,
        init_pdu,
        init_res_var,
        None,
    )
    optimizer = PySaem(nlme_model, obs_df)
    optimizer.run()
    check_surrogate_validity_gp(nlme_model)
    plot_map_estimates(nlme_model)
    plot_weighted_residuals(nlme_model, "iwres")
    plot_weighted_residuals(nlme_model, "pwres")
    plot_weighted_residuals(nlme_model, "npde")



import numpy as np
import torch
import pandas as pd
from vpop_calibration.utils import smoke_test

# Initialize the seeds for all random operators used in the tests
torch.manual_seed(0)

print(f"Initializing test pipeline. {smoke_test=}")



import pickle
import pytest
import os
import pandas as pd
import numpy as np
import itertools

from vpop_calibration import *


@pytest.fixture(scope="session")
def training_data(np_rng):
    # Use this fixture for testing GP training
    patients = {"id": ["p1", "p2"], "k1": [1.0, 2.0]}
    protocol_arms = ["arm-A", "arm-B"]
    outputs = ["s1", "s2"]
    time_steps = np.arange(0, 3.0, 1.0)
    df = pd.DataFrame.from_dict(patients)
    df = df.merge(pd.DataFrame(protocol_arms, columns=["protocol_arm"]), how="cross")
    df = df.merge(pd.DataFrame(outputs, columns=["output_name"]), how="cross")
    df = df.merge(pd.DataFrame(time_steps, columns=["time"]), how="cross")
    df["value"] = np_rng.normal()
    params = ["k1", "time"]
    return df, params


@pytest.fixture(scope="session")
def training_data_bootstrapped(training_data):
    # Same as training data, but incomplete data set
    df, params = training_data
    return (df.sample(frac=0.5), params)


def test_gp_init(training_data, subtests):
    df, params = training_data
    kernel_test = ["RBF", "SMK", "Matern"]
    var_strat_test = ["IMV", "LMCV"]
    mll_test = ["ELBO", "PLL"]
    deep_kernel_test = [True, False]
    training_proportion_test = [1.0, 0.5]
    test_params = itertools.product(
        kernel_test,
        var_strat_test,
        mll_test,
        deep_kernel_test,
        training_proportion_test,
    )
    for i, params_subtest in enumerate(test_params):
        (
            kernel,
            var_strat,
            mll,
            deep_kernel,
            training_proportion,
        ) = params_subtest
        with subtests.test(i=i, msg=f"Testing GP init with {params_subtest}"):
            gp = GP(
                df,
                params,
                var_strat=var_strat,
                mll=mll,
                kernel=kernel,
                deep_kernel=deep_kernel,
                training_proportion=training_proportion,
            )


@pytest.mark.parametrize("mini_batching", [True, False])
@pytest.mark.parametrize("mini_batch_size", [None, 8])
def test_gp_training(training_data, mini_batching, mini_batch_size):
    df, params = training_data
    gp = GP(
        df,
        params,
        nb_training_iter=2,
    )
    gp.train(mini_batching, mini_batch_size)
    gp.eval_perf()


def test_gp_plots(training_data):
    df, params = training_data
    gp = GP(df, params)
    gp.plot_all_solutions("training")
    gp.plot_all_solutions("validation")
    gp.plot_individual_solution(0)
    gp.plot_obs_vs_predicted("training")
    gp.plot_obs_vs_predicted("validation")


def test_gp_incomplete_data(training_data_bootstrapped):
    df, params = training_data_bootstrapped
    gp = GP(df, params, nb_training_iter=2)
    gp.train()
    gp.eval_perf()


@pytest.fixture
def pickle_file():
    model_file = "vpop_calibration/test/gp_model_for_tests.pkl"

    with open(model_file, "wb") as file:
        yield file
    os.remove(model_file)


def test_gp_pickle(training_data, pickle_file):
    df, params = training_data
    gp = GP(df, params)
    pickle.dump(gp, pickle_file)



import pytest
import numpy as np
import pandas as pd

from vpop_calibration.model.data import TrainingDataSet


@pytest.fixture
def training_data(np_rng, include_protocol):
    # Use this fixture for testing GP training
    patients = {"id": ["p1", "p2"], "k1": [1.0, 2.0]}
    protocol_arms = ["arm-A", "arm-B"]
    outputs = ["s1", "s2"]
    time_steps = np.arange(0, 3.0, 1.0)
    df = pd.DataFrame.from_dict(patients)
    if include_protocol:
        df = df.merge(
            pd.DataFrame(protocol_arms, columns=["protocol_arm"]), how="cross"
        )
    df = df.merge(pd.DataFrame(outputs, columns=["output_name"]), how="cross")
    df = df.merge(pd.DataFrame(time_steps, columns=["time"]), how="cross")
    df["value"] = np_rng.normal()
    params = ["k1", "time"]
    return df, params


@pytest.mark.parametrize("training_proportion", [0.5, 1.0])
@pytest.mark.parametrize("log_inputs", [[], ["k1"]])
@pytest.mark.parametrize("log_outputs", [[], ["s1"]])
@pytest.mark.parametrize("data_already_normalized", [False, True])
@pytest.mark.parametrize("include_protocol", [True, False])
def test_loading(
    training_data, training_proportion, log_inputs, log_outputs, data_already_normalized
):
    df, params = training_data
    TrainingDataSet(
        training_df=df,
        descriptors=params,
        log_inputs=log_inputs,
        log_outputs=log_outputs,
        training_proportion=training_proportion,
        data_already_normalized=data_already_normalized,
    )



import pandas as pd
import uuid
import numpy as np
from scipy.stats.qmc import Sobol, scale


def generate_vpop_from_ranges(
    log_nb_individuals: int, param_ranges: dict[str, dict[str, float | bool]]
) -> pd.DataFrame:
    """Generate a vpop of patients from parameter ranges using Sobol sequences

    Args:
        log_nb_individuals (int): The vpop size will be 2^log_nb_individuals
        param_ranges (dict[str, dict[str, float  |  bool]]): One entry for each parameter to be explored
        - `param_name`: {`low`: float, `high`: float, `log`: bool}. Turn `log` to true to define log-scaled ranges

    Returns:
        pd.DataFrame: A set of patients with a generated `id`, and a column per descriptor

    Note:
        This method may be called with an empty dict, to return a list of patient ids.
    """

    nb_individuals = np.power(2, log_nb_individuals)
    params_to_explore = list(param_ranges.keys())
    nb_parameters = len(params_to_explore)
    if nb_parameters != 0:

        # Create a sobol sampler to generate parameter values
        sobol_engine = Sobol(d=nb_parameters, scramble=True)
        sobol_sequence = sobol_engine.random_base2(log_nb_individuals)
        samples = scale(
            sobol_sequence,
            [param_ranges[param_name]["low"] for param_name in params_to_explore],
            [param_ranges[param_name]["high"] for param_name in params_to_explore],
        )

        # Handle log-scaled parameters
        for j, param_name in enumerate(params_to_explore):
            if param_ranges[param_name]["log"] == True:
                samples[:, j] = np.exp(samples[:, j])
        # Create the full data frame of patient descriptors
        patients_df = pd.DataFrame(data=samples, columns=params_to_explore)
    else:
        # No parameter requested, create empty data frame
        patients_df = pd.DataFrame()

    ids = [str(uuid.uuid4()) for _ in range(nb_individuals)]
    patients_df.insert(0, "id", ids)
    return patients_df
